"""
Subnetwork :class:, all stations of a network deployed by one operator
"""
# Standard library modules
import warnings
import logging

# Non-standard modules
from obspy.core.inventory.network import Network as obspy_network

# obsinfo modules
from .network import Network
from .station import Stations
from .operator import Operators
from ..helpers import (str_indent, str_list_str, verify_dict_is_empty,
                       OIDate, Identifiers, Comments)

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Subnetwork(object):
    """
    Equivalent to obspy/StationXML Network

    Attributes:
         campaign_ref (str): campaign reference name:
         operator_ref (str): operator reference name
         network (:class:`Network`): FDSN network
         stations_operators (:class:`.Operators`): default operators for all stations
         stations (list of :class:`.Station`)
         comments (list of str)
         extras (list of str)
    """

    def __init__(self, attributes_dict=None, station_only=False):
        """
        Constructor

        Args:
            attributes_dict (dict or :class:`.ObsMetadata`): dictionary from
                network info file
            station_only (bool): Instructs class Station to create object
                with no instrumentation
        Raises:
            (TypeError): if attributes_dict is empty
        """
        if not attributes_dict:
            msg = 'No network attributes'
            logger.error(msg)
            raise TypeError(msg)

        ref_names = attributes_dict.pop("reference_names", None)
        if ref_names is not None:
            self.campaign_ref = ref_names.get("campaign", None)
            self.operator_ref = ref_names.get("operator", None)
        else:
            self.campaign_ref = None
            self.operator_ref = None
        self.network = Network(attributes_dict.pop("network", None))

        self.operators = Operators(attributes_dict.pop("operators"))
        self.comments = Comments(attributes_dict.pop("comments", None))
        self.extras = [str(k) + ": " + str(v)
                       for k, v in (attributes_dict.pop('extras', {})).items()]
        self.add_extras_to_comments()
        self.stations = Stations(attributes_dict.pop("stations", None),
                                 station_only, self.operators, self.comments)

        verify_dict_is_empty(attributes_dict)


    def __str__(self, indent=0, n_subclasses=0):
        if n_subclasses < 0:
            return f'{type(self)}'
        kwargs = {'indent': 4, 'n_subclasses': n_subclasses-1}
        s = f'{self.__class__.__name__}:\n'
        if self.operator_ref is not None or self.campaign_ref is not None:
            s += f'    references:\n'
            if self.operator_ref is not None:
                s += f'        operator: {self.operator_ref}\n'
            if self.campaign_ref is not None:
                s += f'        campaign: {self.campaign_ref}\n'
        s += f'    network: {self.network.__str__(**kwargs)}\n'
        if len(self.extras) > 0:
            s += f'    extras: {str_list_str(self.extras, **kwargs)}\n'
        if len(self.comments) > 0:
            s += f'    stations_comments: {str_list_str(self.comments, **kwargs)}\n'
        s += f'    stations_operators: {self.operators.__str__(**kwargs)}\n'
        s += f'    stations: {self.stations.__str__(**kwargs)}'
        return str_indent(s, indent)

    def to_obspy(self):
        """
         Convert network object to obspy object

         Returns:
            (:class:~obspy.core.inventory.network.Network): corresponding
                obspy Network
        """
        return obspy_network(
            code=self.network.code,
            description=self.network.name + " - " + self.network.description,
            comments=self.network.comments.to_obspy(),
            start_date=self.network.start_date.to_obspy(),
            end_date=self.network.end_date.to_obspy(),
            restricted_status=self.network.restricted_status,
            identifiers=self.network.identifiers.to_obspy(),
            operators=self.network.operators.to_obspy(),
            source_id=self.network.source_id,
            stations=self.stations.to_obspy(),
            selected_number_of_stations=len(self.stations),
            total_number_of_stations=None,
            alternate_code=None,
            historical_code=None,
            data_availability=None)

    def add_extras_to_comments(self):
        """
        Convert info file extras to XML comments
        """
        if self.extras:
            self.comments += Comments.from_extras(self.extras)



"""
Filter class and subclasses
"""
# Standard library modules
import logging

from .coefficients import Coefficients
from ...helpers import str_indent

logger = logging.getLogger("obsinfo")


class ADConversion(Coefficients):
    """
    ADConversion Filter (Flat Coefficients filter)

    Attributes:
        input_full_scale (float): A/D's input full scale (volts)
        output_full_scale (float): corresponding output full scale (counts)
        delay_samples (float): number of samples that the filter delays an
            impulse
    """

    def __init__(self, attributes_dict, stage_id=-1):
        """
        Constructor

        Args:
            attributes_dict (dict or :class:`.ObsMetadata`): filter attributes
            stage_id (int): id of corresponding stage. Used for reporting only
        """
        logger.debug(f'in {self.__class__.__name__}.__init__()')
        attributes_dict["transfer_function_type"] = 'DIGITAL'
        attributes_dict["numerator_coefficients"] = [1.0]
        attributes_dict["denominator_coefficients"] = []
        # Have to pop these before super, or it will give an error.
        input_full_scale = attributes_dict.pop('input_full_scale', None)
        output_full_scale = attributes_dict.pop('output_full_scale', None)
        super().__init__(attributes_dict, stage_id)
        self.input_full_scale = input_full_scale
        self.output_full_scale = output_full_scale
        self._validate_empty_attributes_dict(attributes_dict)

    def __str__(self, indent=0, n_subclasses=0):
        if n_subclasses < 0:
            return f'{type(self)}'
        s = super().__str__() + '\n'
        s += '    input_full_scale: {self.input_full_scale}\n'
        s += '    output_full_scale: {self.output_full_scale}'
        return str_indent(s, indent)




"""
Write extraction script for LCHEAPO instruments (proprietary to miniseed)
"""
# import obsinfo
from obsinfo.subnetwork import Subnetwork
from ..misc.discoveryfiles import (Datapath)
from ..obsmetadata import (ObsMetadata)
from pathlib import Path
import os.path

SEPARATOR_LINE = "\n# " + 60 * "=" + "\n"


def process_script(network_code,
                   station, station_dir, distrib_dir, input_dir=".",
                   output_dir="miniseed_basic", include_header=True):
    """
    Writes script to transform raw OBS data to miniSEED

    Arguments:
        network_code (str): FDSN network_code
        station (:class: ~obsinfo.Station): the station to process
        station_dir (str): the base directory for the station data
        distrib_dir (str): directory where the lcheapo executables and
                           property files are found
        input_dir (str): directory beneath station_dir for LCHEAPO data
        output_dir (str): directory beneath station_dir for basic miniseed]
        include_header (bool): include the header that sets up paths
                               (should be done once)
    """
    fixed_dir = "lcheapo_fixed"
    s = ""
    if include_header:
        s += __header(station.label)
    s += __setup_variables(distrib_dir, station_dir)
    s += __lcfix_commands(station, input_dir, fixed_dir)
    s += __lc2ms_commands(network_code, station, fixed_dir, output_dir)
    s += __force_quality_commands(output_dir, "D")

    return s


def __header(station_name):

    s = "#!/bin/bash\n"
    s += SEPARATOR_LINE + f'echo "Working on station {station_name}"' + SEPARATOR_LINE
    return s


def __setup_variables(distrib_dir, station_dir):
    """
    distrib_dir: directory containing lcheapo bin/ and config/ directories
                (with lcfix and lc2ms)
    station_dir: base directory for station data files
    """

    s = SEPARATOR_LINE + "# LCHEAPO STEPS" + SEPARATOR_LINE
    s += "#  - Set up paths\n"
    s += f"STATION_DIR={station_dir}\n"
    s += f"LCFIX_EXEC={os.path.join(distrib_dir,'bin','lcfix')}\n"
    s += f"LC2MS_EXEC={os.path.join(distrib_dir,'bin','lc2ms')}\n"
    s += f"LC2MS_CONFIG={os.path.join(distrib_dir,'config','lc2ms.properties')}\n"
    s += f"SDPPROCESS_EXEC={os.path.join(distrib_dir,'bin','sdp-process')}\n"
    s += f"MSMOD_EXEC={os.path.join('/opt/iris','bin','msmod')}\n"
    s += "\n"
    return s


def __lcfix_commands(station, in_path, out_path, in_fnames="*.raw.lch"):

    """
        Write an lc2ms command line

        Inputs:
            in_path:       relative path to directory containing input files
            out_path:      relative path to directory for output files
            in_fnames:     search string for input files within in_path ['*.fix.lch']
         Output:
            string of bash script lines
    """

    s = f'echo "{"-"*60}"\n'
    s += 'echo "Running LCFIX: Fix common LCHEAPO data bugs"\n'
    s += f'echo "{"-"*60}"\n'
    s += f'in_dir="{in_path}"\n'
    s += f'out_dir="{out_path}"\n'

    s += "# - Create output directory\n"
    s += "mkdir $STATION_DIR/$out_dir\n"

    s += "# - Collect input filenames\n"
    s += "command cd $STATION_DIR/$in_dir\n"
    s += f"lchfiles=$(ls {in_fnames})\n"
    s += "command cd -\n"
    s += 'echo "lchfile(s): " $lchfiles\n'

    s += "# - Run executable\n"
    s += '$LCFIX_EXEC $lchfiles -d "$STATION_DIR" -i $in_dir -o $out_dir\n'
    s += "\n"

    return s


def __lc2ms_commands(network_code, station, in_path, out_path,
                     in_fnames="*.fix.lch",
                     out_fnames_model="%E.%S.00.%C.%Y.%D.%T.mseed",
                     force_quality_D=True):
    """
    Write an lc2ms command line

    Arguments:
        station (:class: ~obsinfo.Station): station to process
        in_path (str): relative path to directory containing input files
        in_fnames (str): search string for input files within in_path ['*.fix.lch']
        out_path (str): relative path to directory for output files
        out_fnames_model (str): model for output filenames (should change
            default to '%E.%S.%L.%C.%Y.%D.%T.mseed' once lc2ms handles
            location codes)
        force_quality_D: make a separate call to msmod to force the data
            quality to "D" (should be unecessary once lc2ms is upgraded)
    Returns:
        (str): bash script lines
    """

    network_code = network_code
    station_code = station.label
    obs_type = _get_ref_code(station.instrumentation)
    obs_SN = station.instrumentation.equipment.serial_number
    # CHANNEL CORRESPONDENCES WILL ALLOW THE CHANNEL NAMES TO BE EXPRESSED ON
    # THE COMMAND LINE, WITHOUT USING A DEDICATED CSV FILE
    # channel_corresp = station.instrument.channel_correspondances()

    s = f'echo "{"-"*60}"\n'
    s += 'echo "Running LC2MS: Transform LCHEAPO data to miniseed"\n'
    s += f'echo "{"-"*60}"\n'
    s += f'in_dir="{in_path}"\n'
    s += f'out_dir="{out_path}"\n'

    s += "# - Create output directory\n"
    s += "mkdir $STATION_DIR/$out_dir\n"

    s += "# - Collect input filenames\n"
    s += "command cd $STATION_DIR/$in_dir\n"
    s += f"lchfiles=$(ls {in_fnames})\n"
    s += "command cd -\n"
    s += 'echo "lchfile(s): " $lchfiles\n'

    s += "# - Run executable\n"
    s += '$LC2MS_EXEC $lchfiles -d "$STATION_DIR" -i $in_dir -o $out_dir '
    s += f'-m ":{out_fnames_model}" '
    s += f'--experiment "{network_code}" '
    s += f'--sitename "{station_code}" '
    s += f'--obstype "{obs_type}" '
    s += f'--sernum "{obs_SN}" '
    # s += f'--binding "{channel_corresp}"' '
    s += "-p $LC2MS_CONFIG\n"
    s += "\n"

    return s


def __force_quality_commands(rel_path, quality="D"):
    """ Forces miniseed files to have given quality ('D' by default)
    """
    s = f'echo "{"-"*60}"\n'
    s += f'echo "Forcing data quality to {quality}"\n'
    s += f'echo "{"-"*60}"\n'
    # THE FOLLOWING ASSUMES THAT SDP-PROCESS IS IN OUR PATH, NOT NECESSARILY THE CASE
    s += f'$SDPPROCESS_EXEC -d $STATION_DIR -c="Forcing data quality to {quality}" --cmd="$MSMOD_EXEC --quality {quality} -i {rel_path}/*.mseed"\n'
    s += "\n"
    return s


def _console_script(argv=None):
    """
    Create a bash-script to convert LCHEAPO data to basic miniSEED
    Data should be in station_data_path/{STATION_NAME}/{input_dir}/*.fix.lch

    requires O Dewee program lc2ms, and IRIS program msmod

    """
    from argparse import ArgumentParser

    parser = ArgumentParser(
        prog="obsinfo-make_process_scripts_LC2MS", description=__doc__
    )
    parser.add_argument("subnetwork_file", help="Subnetwork information file")
    parser.add_argument("station_data_path", help="Base path containing stations data")
    parser.add_argument("distrib_path", help="Path to lcheapo software distribution")
    parser.add_argument("-i", "--input_dir", default=".",
                        help="subdirectory of station_data_path/{STATION}/ "
                             "containing input *.raw.lch files")
    parser.add_argument("-o", "--output_dir", default="2_miniseed_basic",
                        help="subdirectory of station_data_path/{STATION}/ "
                             "to put output *.mseed files")
    parser.add_argument("--suffix", default="_LC2MS",
                        help="suffix for script filename")
    parser.add_argument("--append", action="store_true",
                        help="append to existing script file")
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="increase output verbosity")
    parser.add_argument("--no_header", action="store_true",
                        help="do not include a script header")
    parser.add_argument("-q", "--quiet", action="store_true", help="run silently")
    args = parser.parse_args()

    if not args.quiet:
        print("Creating LC2MS process scripts, ", end="", flush=True)

    # READ IN NETWORK INFORMATION
    dp = Datapath()
    if args.verbose:
        print(f'Reading network file: {args.network_file}')
    args.subnetwork_file = str(Path(os.getcwd()).joinpath(args.subnetwork_file))
    info_dict = ObsMetadata.read_info_file(args.subnetwork_file, dp, False)
    subnet_dict = info_dict.get('subnetwork', None)
    if not subnet_dict:
        return
    if args.verbose:
        print(f'Processing subnetwork file: {args.subnetwork_file}')
    subnetwork = Subnetwork(ObsMetadata(subnet_dict))

    if not args.quiet:
        print(f"network {network.fdsn_code}, subnetwork stations", end="", flush=True)
        if args.verbose:
            print("")

    first_time = True
    for station in subnetwork.stations:
        if not args.quiet:
            if args.verbose:
                print(f"\t{station.label}", end="")
            else:
                if first_time:
                    print(f"{station.label}", end="", flush=True)
                else:
                    print(f", {station.label}", end="", flush=True)
        station_dir = os.path.join(args.station_data_path, station.label)
        script = process_script(
            subnetwork.fdsn_code,
            station,
            station_dir,
            args.distrib_path,
            input_dir=args.input_dir,
            output_dir=args.output_dir,
            include_header=not args.no_header,
        )
        fname = "process_" + station.label + args.suffix + ".sh"
        if args.verbose:
            print(f" ... writing file {fname}", flush=True)
        if args.append:
            write_mode = "a"
        else:
            write_mode = "w"
        with open(fname, write_mode) as f:
            # f.write('#'+'='*60 + '\n')
            f.write(script)
            f.close()
        first_time = False
    if not args.verbose and not args.quiet:
        print("")


def _get_ref_code(inst):
    """
    Returns the LCHEAPO reference code corresponding to the instrumentation

    Arguments:
        inst (:class: ~obsinfo.Instrumentation)
    """
    sps = 50
    ch = inst.channels
    if len(ch) == 2:
        if (ch[0].channel_code(sps) == 'SH3' and ch[1].channel_code(sps) == 'BDH'):
            return 'SPOBS1'
    elif len(ch) == 4:
        if   (ch[0].channel_code(sps) == 'BDH'
              and ch[1].channel_code(sps) == 'SH2'
              and ch[2].channel_code(sps) == 'SH1'
              and ch[3].channel_code(sps) == 'SH3'):
            return 'SPOBS2'
        elif (ch[0].channel_code(sps) == 'BH2'
              and ch[1].channel_code(sps) == 'BH1'
              and ch[2].channel_code(sps) == 'BHZ'
              and ch[3].channel_code(sps) == 'BDH'):
            return 'BBOBS1'
        elif (ch[0].channel_code(sps) == 'BDH'
              and ch[1].channel_code(sps) == 'BDH'
              and ch[2].channel_code(sps) == 'BDH'
              and ch[3].channel_code(sps) == 'BDH'):
            return 'HYDROCT'
    return "UNKNOWN"



"""
Digital filter (subclass of Coefficents, which is subclass of Filter)
"""
# Standard library modules
import logging

from .coefficients import Coefficients
from ...helpers import str_indent

logger = logging.getLogger("obsinfo")


class Digital(Coefficients):
    """
    Digital Filter (Flat Coefficients filter)
    """
    def __init__(self, attributes_dict, stage_id=-1):
        """
        Constructor

        Args:
            attributes_dict (dict or :class:`.ObsMetadata`): filter attributes
            stage_id (int): id of corresponding stage. Used for reporting only
        """
        logger.debug(f'in {self.__class__.__name__}.__init__()')
        attributes_dict["transfer_function_type"] = 'DIGITAL'
        attributes_dict["numerator_coefficients"] = [1.0]
        attributes_dict["denominator_coefficients"] = []
        super().__init__(attributes_dict, stage_id)
        self._validate_empty_attributes_dict(attributes_dict)

    def __repr__(self):
        return 'Digital(attributes_dict, stage_id)'

    def __str__(self, indent=0, n_subclasses=0):
        if n_subclasses < 0:
            return f'{type(self)}'
        s = super().__str__()
        return str_indent(s, indent)




"""
Subnetwork and Station(s) classes and their helpers
"""
# The noqa comments disables flake8's "imported but not used" error
from .subnetwork import Subnetwork               # noqa: F401
from .network import Network                     # noqa: F401
from .station import Station, Stations           # noqa: F401
from .operator import Operator, Operators        # noqa: F401
from .site import Site                           # noqa: F401



"""
Location classe
"""
# Standard library modules
import warnings
import numpy as np
import logging

from .float_with_uncert import FloatWithUncert
from ..obsmetadata import ObsMetadata
from .functions import str_indent
from .obsinfo_class_list import ObsinfoClassList

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")

class Locations(ObsinfoClassList):
    def __init__(self, inp):
        super().__init__(inp, Location)

    def get_by_loc_code(self, loc_code):
        for loc in self:
            if loc.code == loc_code:
                return loc
        return None

    @classmethod
    def from_locations_dict(cls, locations_dict):
        """
        Create from a locations dict (key=location_code, value=location_dict)
        """
        return cls([Location(v, c) for c, v in locations_dict.items()])

class Location(object):
    """
    Location class.

    Attributes:
        latitude (float): station latitude (degrees N)
        longitude (float): station longitude (degrees E)
        elevation (float): station elevation (meters above sea level)
        uncertainties_m (dict): 'lat', 'lon', 'elev' in METERS
        geology (str): site geology
        vault (str): vault type
        depth_m (float): depth of station beneath surface (meters)
        water_level (float): elevation of water_level (non-zero for lakes)
        localisation_method (str): method used to determine position
        obspy_latitude: latitude as an *obspy* object
        obspy_longitude: longitude as an *obspy* object

    """

    def __init__(self, attributes_dict, code:str=None):
        """
        Create Location object and assign attributes from attributes_dict.
        Validate required location attributes exist
        Convert to obspy longitude and latitude

        Args:
            attributes_dict (dict or :class:`ObsMetadata`): location
                information
            code (str): location code, error if 'code' in attributes_dict
        """
        attributes_dict = ObsMetadata(attributes_dict)
        if code is not None:
            if 'code' in attributes_dict:
                raise ValueError('provided location code as argument AND dict key')
            else:
                attributes_dict['code'] = code

        position = attributes_dict.pop('position', None)
        if position is None:
            msg = 'No position in location'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)
        self._lat = position.pop('lat')
        self._lon = position.pop('lon')
        self._elev = position.pop('elev')
        self._position = position  # for __repr__()
        self.code = attributes_dict.pop('code', None)

        # Get base-config elements
        base_dict = attributes_dict.get_configured_modified_base()
        self.geology = base_dict.get('geology', None)
        self.vault = base_dict.get('vault', None)
        self.depth_m = base_dict.get('depth.m', None)
        self.water_level = base_dict.get('water_level.m', None)
        self._uncert_m = base_dict.get('uncertainties.m', None)
        self._measurement_method = base_dict.get('measurement_method', None)

        self.latitude = FloatWithUncert({
            "value": self._lat,
            "uncertainty": self._uncert('lat'),
            "measurement_method": self._measurement_method})
        self.longitude = FloatWithUncert({
            "value": self._lon,
            "uncertainty": self._uncert('lon'),
            "measurement_method": self._measurement_method})
        self.elevation = FloatWithUncert({
            "value": self._elev,
            "uncertainty": self._uncert('elev'),
            "measurement_method": self._measurement_method})

    def __repr__(self):
        args = []
        args.append(f"'position': {self._position}")
        args.append(f"'uncertainties.m': {self._uncert_m}")
        args.append(f"'measurement_method': '{self._measurement_method}'")
        if not self.geology == 'unknown':
            args.append(f"'geology': '{self.geology}'")
        if self.vault:
            args.append(f"'vault'='{self.vault}'")
        if self.depth_m is not None:
            args.append(f"'depth_m'={self.depth_m:g}")
        s = 'Location({' + ', '.join(args) + '})'
        return s

    def __str__(self, indent=0, n_subclasses=0):
        s = f'Location:\n'
        s += f'    code: {self.code}\n'
        s += f'    latitude: {self.latitude}\n'
        s += f'    longitude: {self.longitude}\n'
        s += f'    elevation: {self.elevation}'
        if not self.geology == 'unknown':
            s += f'\n    geology: {self.geology}'
        if self.vault:
            s += f'\n    vault: {self.vault}'
        if self.depth_m is not None:
            s += f'\n    depth: {self.depth_m:g}m'
        if self.water_level is not None:
            s += f'\n    water_level: {self.water_level:g}m'
        if self._measurement_method:
            s += f'\n    measurement_method: "{self._measurement_method}"'
        return str_indent(s, indent)

    def _uncert(self, which):
        """
        Returns uncertainty in appropriate units

        Args:
            which (str): must be 'lat'', 'lon', or 'elev'
        """
        if which == 'elev':
            return self._uncert_m.get('elev', None)
        elif which == 'lat':
            uncert_m = self._uncert_m.get('lat', None) \
                if self._uncert_m else None
            if uncert_m is None:
                return None
            else:
                meters_per_degree_lat = 1852.0 * 60.0
                lat_uncert = uncert_m / meters_per_degree_lat
                # cut off extraneous digits
                lat_uncert = float("{:.3g}".format(lat_uncert))
                return lat_uncert
        elif which == 'lon':
            uncert_m = self._uncert_m.get('lon', None) \
                if self._uncert_m else None
            if uncert_m is None or abs(self._lat) == 90:
                return None
            else:
                m_per_deg_lon = (1852.0 * 60.0 * np.cos(np.radians(self._lat)))
                lon_uncert = uncert_m / m_per_deg_lon
                lon_uncert = float("{:.3g}".format(lon_uncert))
                return lon_uncert
        else:
            raise ValueError('Illegal position type "{which}", should be lat, '
                             'lon or elev')



"""
Instrument and Operator classes
"""
# Standard library modules
import warnings
import logging
from copy import deepcopy

# Non-standard modules
from obspy.core.inventory.response import (Response, InstrumentSensitivity
                                           as obspy_Sensitivity)

# obsinfo modules
from .instrument_component import (InstrumentComponent,
                                   Datalogger, Sensor, Preamplifier)

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Instrument(object):    # Was a Channel subclass, but don't see why
    """
    An instrument is an ensemble of a sensor, a datalogger and possibly a
    preamplifier. It also includes a selected configuration for each one of
    these instrument components.

    Attributes:
        datalogger (:class:`Datalogger`)
        sensor: (:class:`Sensor`)
        preamplifier: (:class:`Preamplifier`)
        sample_rate (float): from datalogger sample rate
        correction (float): from datalogger delay correction
        seed_band_base_code (str): from sensor band base code
        seed_instrument_code (str): from sensor instrument code
    """

    def __init__(self, attributes_dict, modifs={}):
        """
        Constructor

        Args:
            attributes_dict (dict or :class:`.ObsMetadata`): instrument attributes_dict
            modifs (dict or :class:`ObsMetadata`): modifications passed
                down from the Instrumentation level (including channel selection)
        """
        # For __repr__()
        if modifs:
            self.inputs = {'modifs': "<ObsMetadata>"}
        else:
            self.inputs = {'modifs': {}}
        
        self.correction = None
        self.delay = None

        if not attributes_dict:
            msg = 'No instrument attributes_dict'
            logger.error(msg)
            raise ValueError(msg)

        # Create the three InstrumentComponents
        self.datalogger = Datalogger(attributes_dict.pop('datalogger', {}),
                                     modifs.get('datalogger', {}))
        self.sensor = Sensor(attributes_dict.pop('sensor', {}),
                             modifs.get('sensor', {}))
        if not 'preamplifier' in attributes_dict:
            self.preamplifier = None
        else:
            self.preamplifier = Preamplifier(attributes_dict.pop('preamplifier', {}),
                                             modifs.get('preamplifier', {}))
        # Combine the InstrumentComponents' response stages
        self._combine_stages()
        # Validate inputs and outputs and correct delay
        self._integrate_stages()
        
        if attributes_dict:
            raise ValueError(f'attributes_dict is not empty after reading: {attributes_dict=}')
        # # if self.stages is not None:
        # #     self.obspy_stages = [x.to_obspy() for x in self.stages]
        # # else:
        # #     self.obspy_stages = None
        # self.obspy_response = self.to_obspy()

    def __repr__(self):
        s =   'Instrument(attributes_dict = <ObsMetadata>\n',
        s += f'           modifs={self.inputs["modifs"]}'
        return s

    def __str__(self, indent=0, n_subclasses=0):
        if n_subclasses < 0:
            return self.__class__.__name__

        kwargs = dict(indent=4, n_subclasses=n_subclasses-1)
        s =  'Instrument:\n'
        s += f'  sensor: {self.sensor.equipment.__str__(**kwargs)}\n'
        if self.preamplifier is not None:
            s += f'  preamplifier: {self.preamplifier.equipment.description}\n'
        else:
            s += f'  preamplifier: None\n'
        s += f'  datalogger: {self.datalogger.equipment.__str__(**kwargs)}\n'
        s += f'  stages: {self.stages.__str__(**kwargs)}\n'
        s += f'  sample_rate: {self.sample_rate}\n'
        s += f'  delay: {self.delay}\n'
        if self.correction is None:
            corrections = [x.correction for x in self.stages if x.correction is not None and not x.correction == 0]
            s += '  correction: {} (sum of non-zero corrections in {} stages)\n'.format(
                sum(corrections), len(corrections))
        else:
            s += f'  correction: {self.correction}\n'
        s += f'  seed_band_base_code: {self.seed_band_base_code}\n'
        s += f'  seed_instrument_code: {self.seed_instrument_code}'
        return s

    def to_obspy(self):
        """
        Return equivalent obspy class

        Returns:
            ():class:`obspy.core.inventory.response.Response`)
        """
        sensitivity = self._calc_sensitivity()
        if self.stages is not None:
            obspy_stages = [x.to_obspy() for x in self.stages]
        else:
            obspy_stages = None
        return Response(resource_id=None,
                        instrument_sensitivity=sensitivity,
                        instrument_polynomial=None,
                        response_stages=obspy_stages)

    def _combine_stages(self):
        """
        Adds all response stages as obsinfo and obpsy objects and renumbers
        them

        Returns stages as a Stages object
        """
        if self.sensor.stages is not None:
            self.stages = self.sensor.stages
        else:
            self.stages = Stages()

        if self.preamplifier is not None:
            if self.preamplifier.stages is not None:
                self.stages += self.preamplifier.stages

        if self.datalogger.stages is not None:
            self.stages += self.datalogger.stages

        # Order the stage_sequence_numbers
        i = 1
        for s in self.stages:
            s.stage_sequence_number = i
            i += 1

    def _integrate_stages(self):
        """
        Integrates the stages with one another

        1) Renumber stages sequentially
        2) Verify/set units and sample rates
        3) Assure same frequency is used for consecutive PZ filters
        4) Calculate global polarity of the whole set of response stages
        5) Set global response delay correction
        6) Validate sample_rate expressed in datalogger component is equal to
           global response sample rate
        """
        if self.stages is None or len(self.stages) == 0:
            self.polarity = None
            return

        polarity = 1
        prev_pz_norm_freq = None
        prev_stage = None
        for stage in self.stages:
            if prev_stage is not None:
                prev_ssn = prev_stage.stage_sequence_number
                this_ssn = stage.stage_sequence_number

                # 2a) Verify continuity of units
                if prev_stage.output_units != stage.input_units:
                    msg = "Stage {} and {} units don't match".format(
                        prev_ssn, this_ssn)
                    warnings.warn(msg)
                    logger.error(msg)
                    raise ValueError(msg)

                # 2b) Verify/set continuity of sample rate
                if prev_stage.output_sample_rate:
                    if stage.input_sample_rate:
                        if prev_stage.output_sample_rate != stage_input_sample_rate:
                            msg = ("stage {this_ssn} input sample rate "
                                   "doesn't match previous stage's output "
                                   "sample rate ({stage.input_sample_rate} "
                                   "!= {prev_stage.output_sample_rate)}")
                            warnings.warn(msg)
                            logger.error(msg)
                            raise ValueError(msg)
                    else:
                        stage.input_sample_rate = prev_stage.output_sample_rate
                        
                # 3) Check that all PZ stages have the same normalization frequency.
                if stage.filter.type == 'PolesZeros':
                    if prev_pz_norm_freq is not None:
                        if (prev_pz_norm_freq != stage.filter.normalization_frequency
                            and prev_pz_norm_freq != 0):
                            msg = ("Normalization frequencies for PZ stages "
                                   f"{prev_ssn} and {this_ssn} don't match "
                                   f"({prev_pz_norm_freq} "
                                   f"!= {stage.filter.normalization_frequency})")
                            warnings.warn(msg)
                            logger.warning(msg)
                    prev_pz_norm_freq = stage.filter.normalization_frequency

            # 4) Calculate/verify delay and correction
            if self.delay is None:
                self.delay = stage.delay
            elif stage.delay is not None:
                self.delay += stage.delay
            if self.correction is None and stage.delay is not None:
                stage.correction = stage.delay

            # 5) Calculate global polarity
            if not stage.polarity:  # default polarity is positive
                stage.polarity = 1
            polarity *= stage.polarity

            # Save previous stage for comparison
            prev_stage = deepcopy(stage)
            first_stage = False

        if self.correction is not None:
            self.stages[-1].correction = self.correction

        # Check global output sample rate
        if not self.stages[-1].output_sample_rate == self.sample_rate:
            msg = ('Declared sample rate != calculated sample rate '
                   f'({self.sample_rate} != {self.stages[-1].output_sample_rate})')
            logger.error(msg)
            raise ValueError(msg)

        # Set global response attributes
        self.polarity = polarity

    def _calc_sensitivity(self):
        """
        Calculates sensitivity
        Based on ..misc.obspy_routines.response_with_sensitivity
        """
        response_stg = self.stages
        gain_prod = 1.
        if response_stg is None:
            iu = "None"
            ou = "None"
            iud = "None"
            oud = "None"
            gain_freq = 0
        else:
            iu = response_stg[0].input_units
            ou = response_stg[-1].output_units
            iud = response_stg[0].input_units_description
            oud = response_stg[-1].output_units_description
            # gain_frequency could be provided, according to StationXML, but we
            # assume it's equal to the gain frequency of first stage
            gain_freq = response_stg[0].gain_frequency,

            if "PA" in iu.upper():
                # MAKE OBSPY THINK ITS M/S TO CORRECTLY CALCULATE SENSITIVITY
                sens_iu = "M/S"
            else:
                sens_iu = iu
            for stage in response_stg:
                gain_prod *= stage.gain

        sensitivity = obspy_Sensitivity(gain_prod, gain_freq,
                                        input_units=sens_iu, output_units=ou,
                                        input_units_description=iud,
                                        output_units_description=oud)
        sensitivity.iu = iu
        return sensitivity

    def get_response_stage(self, num):
        """
        Returns the response stage in a given position

        Args:
            num (int): stage number, starting with zero and ordered from
                sensor to datalogger
        """
        # All response stages are at the instrument_component level
        stages = self.stages
        assert(num <= stages[-1].stage_sequence_number), \
            'response stage out of range: {num}'
        return stages[num]

    @property
    def equipment_datalogger(self):
        return self.datalogger.equipment

    @property
    def equipment_sensor(self):
        return self.sensor.equipment

    @property
    def equipment_preamplifier(self):
        return self.preamplifier.equipment

    @property
    def sample_rate(self):
        return self.datalogger.sample_rate

    @property
    def seed_band_base_code(self):
        return self.sensor.seed_band_base_code

    @property
    def seed_instrument_code(self):
        return self.sensor.seed_instrument_code



import gitlab
# import urllib
# import json
import base64
import re
# import os
# from pathlib import Path
from urllib import parse as urlparse
# from urllib.parse import unquote
# from urllib.request import urlopen
import logging

# obsinfo
from .configuration import ObsinfoConfiguration

# private token or personal token authentication

logger = logging.getLogger("obsinfo")


class gitLabFile(object):
    """
    Provide the methods to use the gitlab API to read a remote file and decode it

    Attributes:
        *None*
    """
    @staticmethod
    def get_gitlab_file(uri, read_file=True):
        """
        Get the file content pointed at by uri and decode it.

        Uses b64 first to get the remote file and convert to a byte string,
        and then utf-8 to convert to regular string.

        Args:
            uri (string or path-like): uri to read
        :param read_file (bool): If true, reads the content. If not, simply
            checks if the file exists

        Returns:
            (str): read content

        Raises: FileNotFoundError, ValueError
        """
        # Use configuration ifnormation
        gitlab_repository = ObsinfoConfiguration.Instance().gitlab_repository
        project_path = ObsinfoConfiguration.Instance().gitlab_project
        obsinfo_branch = ObsinfoConfiguration.Instance().obsinfo_branch

        if not gitlab_repository or not project_path:
            msg = "One or several environment variables, OBSINFO_GITLAB_REPOSITORY or \
                     OBSINFO_PROJECTPATH in the configuration file ~/.obsinforc, are missing"
            logger.error(msg)
            raise ValueError(msg)

        with gitlab.Gitlab(gitlab_repository) as gl:

            project = gl.projects.get(project_path)
            if not project:
                msg = f'Project not found in repository: {project_path}'
                logger.error(msg)
                raise FileNotFoundError(msg)

            if urlparse.urlsplit(uri).scheme:  # Sometime it comes with scheme, sometimes not
                uri = urlparse.urlsplit(uri).path
                pattern = "/"
            else:
                # Remove elements from path
                pat1 = re.compile("http://")
                pat2 = re.compile("https://")
                if pat1.match(gitlab_repository):
                    gitlab_repository = pat1.sub("", gitlab_repository)
                if pat2.match(gitlab_repository):
                    gitlab_repository = pat2.sub("", gitlab_repository)

                pattern = ("/" if gitlab_repository[0] != "/" else "") + gitlab_repository

                if gitlab_repository[-1] != "/":
                    pattern += "/"

            # we assume project_path has no leading slash
            pattern += project_path
            if project_path[-1] != "/":
                pattern += "/"

            # remove everything but path relative to project path
            uri = re.sub(pattern, "", uri)

            try:
                f = project.files.get(file_path=uri, ref=obsinfo_branch)
                if not read_file:
                    return True
            except gitlab.exceptions.GitlabOperationError:
                if read_file:
                    msg = f'Error reading remote (gitlab) file: {uri}'
                    print(msg)
                    raise FileNotFoundError(msg)
                else:
                    return None
            # get the decoded content. Two decodes are necessary, from 8character string to bytes and
            # from bytes to a python string using utf-8
            bytecontent = base64.b64decode(f.content)
            ret = bytecontent.decode("utf-8")

        return ret

    @staticmethod
    def isRemote(file):
        """
        Checks if scheme means file is remote.

        :param file: filename to be checked, with complete uri
        :type file: str
        :returns: boolean. True if remote, False otherwise
        """
        return re.match('^http:', file) or re.match('^https:', file)



"""
Stages and Stage classes
"""
# Standard library modules
import warnings
import re
import logging

# Non-standard modules
from obspy.core.inventory.response import (
    PolesZerosResponseStage, FIRResponseStage,
    CoefficientsTypeResponseStage, ResponseListResponseStage,
    PolynomialResponseStage)
import obspy.core.util.obspy_types as obspy_types

# Local modules
from ..obsmetadata import ObsMetadata
from .filter import (Filter, PolesZeros, FIR, Coefficients, ResponseList,
                     Analog, Digital, ADConversion, Polynomial)
from ..helpers import str_indent, OIDates

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Stage(object):
    """
    Stage is a discrete portion of the response of the instrument

    Attributes:
        name (str): name of stage, if any
        description (str): description of stage
        input_units (str): validated in schema
        input_units_description (str)
        output_units (str): validated in schema
        output_units_description (str)
        gain (float): value of gain
        gain_frequency (float): frequency at which gain is measured
        filter (object of :class:`Filter`)
        stage_sequence_number: sequence number in total response, assigned
            later
        input_sample_rate (float): input sample rate in sps
        delay (float): delay in seconds of stage. If not present, will be
            calculated from delay from digital stages
        decimation_factor (float): decimation factor of stage
        correction (float) : delay correction. Calculated from instrument
            delay correction
        polarity (str, either "+" or "-"): whether stage changes polarity
        resource_id (str)
        instrument_sensitivity (float): Not used, set to None. Sensitivity
            is calculated for the whole response.

    """
    def __init__(self, attributes_dict, higher_modifs=ObsMetadata({}),
                 correction=None, sequence_number=-1,
                 ext_config_name=None):
        """
        Args:
            attributes_dict (dict or :class:`ObsMetadata`): attributes of
                component
            higher_modifs (dict or :class:`ObsMetadata`):
                response modifications from higher level 
                (key = regex for stage #, value = modification dictionary)
            correction (float): used only for datalogger, it's the delay
                correction of the whole instrument
            sequence_number (int): sequence number, starting at 1. First
                assigned within component, then for the whole instrument
                response
           ext_config_name (str): higher-level configuration definition (overrides
                whatever is in self)
        """
        if attributes_dict is None:
            return None
        
        # input parameters, for __repr__
        self.inputs = {'attributes_dict': '<ObsMetadata>',
                       'higher_modifs': '<ObsMetadata>',
                       'correction': correction,
                       'sequence_number': sequence_number,
                       'ext_config_name': ext_config_name}
        if not higher_modifs:
                    self.inputs['higher_modifs'] = '{}'

        if not isinstance(attributes_dict, ObsMetadata):
            attributes_dict = ObsMetadata(attributes_dict)
        if not isinstance(higher_modifs, ObsMetadata):
            higher_modifs = ObsMetadata(higher_modifs)

        # put appropriate "stage_modifications" in "modifications"
        m = self._get_stage_modifications(
            attributes_dict.pop('stage_modifications', {}), sequence_number)
        if "modifications" in attributes_dict:
            attributes_dict['modifications'].safe_update(m)
        elif not m == {}:
            attributes_dict['modifications'] = m

        hm = self._get_stage_modifications(
            higher_modifs.pop('stage_modifications', {}), sequence_number)
        if "modifications" in higher_modifs:
            higher_modifs['modifications'].safe_update(hm)
        elif not hm == {}:
            higher_modifs['modifications'] = hm

        base_dict = attributes_dict.get_configured_modified_base(higher_modifs)

        self.name = base_dict.pop('name', None)  
        self.configuration = base_dict.pop('configuration', None)
        # CalibrationDates belongs in equipment, but is more logical to state
        # in stage, rely on equipment to look for it.    
        self.calibration_dates = OIDates(base_dict.pop('calibration_dates', []) )     
        self.configuration_description = base_dict.pop('configuration_description',
                                                       self.configuration)      
        self.description = base_dict.pop('description', '')
        if self.configuration_description is not None:
            if self.name is None:
                self.name = f'[config: {self.configuration_description}]'
            else:
                self.name += f' [config: {self.configuration_description}]'

        gd = base_dict.pop('gain', None)
        if gd is None:
            msg = f'No gain specified in stage {self.name}'
            logger.error(msg)
            raise TypeError(msg)
        self.gain = gd.get('value', 1.0)
        self.gain_frequency = gd.get('frequency', 0.0)
        self.stage_sequence_number = sequence_number


        self.filter = Filter.construct(base_dict.pop('filter'),
                                       self.stage_sequence_number,
                                       self.name,
                                       self.gain_frequency)
        if not self.filter:
            msg = f'No filter in stage {self.name}'
            logger.error(msg)
            raise TypeError(msg)

        x = base_dict.pop('input_units', None)
        if x:
            self.input_units = x.get('name', None)
            self.input_units_description = x.get('description', None)

        x = base_dict.pop('output_units', None)
        if x:
            self.output_units = x.get('name', None)
            self.output_units_description = x.get('description', None)

        self.input_sample_rate = base_dict.pop('input_sample_rate', None)

        # Set an unconfigured delay to None so that it can be
        # calculated in self.calculate_delay(), once input_sample_rate
        # is known
        self._delay = base_dict.pop('delay', None)
        self.decimation_factor = base_dict.pop('decimation_factor', 1)
        self.correction = correction

        # default polarity is positive
        self.polarity = base_dict.pop('polarity', 1)
        self.resource_id = base_dict.pop('resource_id', None)
        self.instrument_sensitivity = None
        # Instrument sensitivity will be calculated using obspy and stored in
        # first stage
        if len(base_dict) > 0:
            raise ValueError('base_dict has remaing fields: {}'
                             .format([x for x in base_dict.keys()]))

    @property
    def output_sample_rate(self):
        """
        Output sample rate is not specified but calculated from
        input sample rate and decimation factor
        """
        if self.input_sample_rate and self.decimation_factor:
            return self.input_sample_rate / self.decimation_factor
        else:
            return None

    @property
    def delay(self):
        """
        Calculates delay in seconds

        Delay is a function of filter offset for digital filters if not
        specified in info file
        """
        filter_delay = None
        
        # Check for "impossible" error
        if self.filter.delay_seconds is not None and self.filter.delay_samples is not None:
            ValueError('delay_seconds and delay_samples both set for {self.filter.type} filter')

        # Calculate delay based on filter
        if self.filter.delay_seconds is not None:
            filter_delay = self.filter.delay_seconds
        elif self.filter.delay_samples is not None:
            if not self.input_sample_rate:
                # Delay is already none, leave it like that
                msg = ('Cannot calculate delay from delay_samples '
                       f'because stage {self.name} has no input_sample_rate')
                logger.warning(msg)
                warnings.warn(msg)
            else:
                filter_delay = self.filter.delay_samples / self.input_sample_rate

        if self._delay is None:
            return filter_delay 
        
        # Compare stage-given and filter-based delays
        if filter_delay is not None and not self._delay == filter_delay:
            msg = ("stage-set delay does not equal filter delay"
                   f" ({self._delay} != {filter_delay}): ignoring filter "
                   "delay")
            warning.warn(msg)
            logger.warning(msg)
        return self._delay
        
    def __repr__(self):
        return (f"Stage({self.inputs['attributes_dict']}, "
                f"{self.inputs['higher_modifs']}, "
                f"{self.inputs['correction']}, "
                f"{self.inputs['sequence_number']}, "
                f"{self.inputs['ext_config_name']})")

    def __str__(self, indent=0, n_subclasses=0):
        if n_subclasses < 0:
            return f'{type(self)}'
        kwargs = {'indent': 4, 'n_subclasses': n_subclasses-1}
        s = f'Stage:\n'
        s += f'    name: {self.name}\n'
        s += f'    description: {self.description}\n'
        s += f'    input_units: {self.input_units}\n'
        s += f'    output_units: {self.output_units}\n'
        s += f'    gain: {self.gain}\n'
        s += f'    gain_frequency: {self.gain_frequency:g}\n'
        s += f'    filter: {self.filter.__str__(**kwargs)}\n'
        if not self.stage_sequence_number == -1:
            s += f'    stage_sequence_number: {self.stage_sequence_number}\n'
        if self.calibration_dates is not None:
            s += f'    calibration_dates: {self.calibration_dates.__str__(**kwargs)}\n'
        if self.input_units_description:
            s += f'    input_units_description: {self.input_units_description}\n'
        if self.output_units_description:
            s += f'    output_units_description: {self.output_units_description}\n'
        if self.input_sample_rate:
            s += f'    input_sample_rate: {self.input_sample_rate}\n'

        s += f'    decimation_factor: {self.decimation_factor}\n'
        s += f'    delay: {self.delay}\n'
        s += f'    correction: {self.correction}'

        return str_indent(s, indent)

    def to_obspy(self):
        """
        Return equivalent *obspy.core.inventory.response* classes stages:

        Possible stage classes:

           * PolesZerosResponseStage
           * FIRResponseStage
           * CoefficientsTypeResponseStage
           * ResponseListResponseStage
           * Response

        :returns: object of one the above classes
        """
        # Arguments for all
        args = (self.stage_sequence_number, self.gain, self.gain_frequency,
                self.input_units, self.output_units)
        kwargs = dict(name=self.name,
                      input_units_description=self.input_units_description,
                      output_units_description=self.output_units_description,
                      description=self.description,
                      decimation_input_sample_rate=self.input_sample_rate,
                      decimation_factor=self.decimation_factor,
                      decimation_delay=self.delay,
                      decimation_correction=self.correction,
                      resource_id=self.resource_id,
                      resource_id2=self.filter.resource_id)

        # Filter-dependent arguments
        filt = self.filter
        if isinstance(filt, Analog):  # subclass of PolesZeros, so test first
            if not filt.normalization_frequency:
                filt.normalization_frequency = self.gain_frequency
            obspy_stage = PolesZerosResponseStage(
                *args, **kwargs,
                pz_transfer_function_type=filt.transfer_function_type,
                normalization_frequency=filt.normalization_frequency,
                zeros=filt.zeros,
                poles=filt.poles,
                normalization_factor=filt.normalization_factor)

        elif isinstance(filt, PolesZeros):
            if not filt.normalization_frequency:
                filt.normalization_frequency = self.gain_frequency
            obspy_stage = PolesZerosResponseStage(
                *args, **kwargs,
                pz_transfer_function_type=filt.transfer_function_type,
                normalization_frequency=filt.normalization_frequency,
                zeros=[obspy_types.ComplexWithUncertainties(t)
                       for t in filt.zeros],
                poles=[obspy_types.ComplexWithUncertainties(t)
                       for t in filt.poles],
                normalization_factor=filt.normalization_factor)

        elif isinstance(filt, FIR):
            obspy_stage = FIRResponseStage(
                *args, **kwargs,
                symmetry=filt.symmetry,
                coefficients=[obspy_types.FloatWithUncertaintiesAndUnit(
                    c / filt.coefficient_divisor) for c in filt.coefficients])
        # subclasses of Coefficients, so test first
        elif (isinstance(filt, Digital) or isinstance(filt, ADConversion)):
            obspy_stage = CoefficientsTypeResponseStage(
                *args, **kwargs,
                cf_transfer_function_type=filt.transfer_function_type,
                numerator=filt.numerator_coefficients,
                denominator=filt.denominator_coefficients)
        elif isinstance(filt, Coefficients):
            obspy_stage = CoefficientsTypeResponseStage(
                *args, **kwargs,
                cf_transfer_function_type=filt.transfer_function_type,
                numerator=[obspy_types.FloatWithUncertaintiesAndUnit(
                           n, lower_uncertainty=0.0, upper_uncertainty=0.0)
                           for n in filt.numerator_coefficients],
                denominator=[obspy_types.FloatWithUncertaintiesAndUnit(
                             n, lower_uncertainty=0.0, upper_uncertainty=0.0)
                             for n in filt.denominator_coefficients]
                )
        elif isinstance(filt, ResponseList):
            obspy_stage = ResponseListResponseStage(
                *args, **kwargs,
                response_list_elements=[
                    obspy_types.ResponseListElement(x[0], x[1], x[2])
                    for x in filt.elements]
                )
        elif isinstance(filt, Polynomial):
            obspy_stage = PolynomialResponseStage(
                *args, **kwargs,
                approximation_type=filt.approximation_type,
                frequency_lower_bound=filt.frequency_lower_bound,
                frequency_upper_bound=filt.frequency_upper_bound,
                approximation_lower_bound=filt.approximation_lower_bound,
                approximation_upper_bound=filt.approximation_upper_bound,
                maximum_error=filt.aximum_error,
                coefficients=filt.coefficients
                )
        else:
            msg = 'Unhandled response stage type in stage '\
                  f'#{self.stage_sequence_number}: "{filt.type}"'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)

        return obspy_stage
    
    @staticmethod
    def _get_stage_modifications(resp_modifs, sequence_number):
        """
        Select which channel modifications specified at station level apply
        to a given stage,  with the stage number (WITHIN an instrument
        component) as key code

        Args:
            resp_modifs (dict or :class:`.ObsMetadata`): response modifications:
                key = regex stage number e.g.: "*", "[1,2,4]" or "[1-3]"
                value = modification dict
            sequence_number (int): stage sequence number (within component)
        """
        stage_code = str(sequence_number)
        default_dict = range_dict = {}
        modif = resp_modifs.get(str(stage_code), {})

        match_key = None
        for k, v in resp_modifs.items():
            if k[0] == "*":
                default_dict = v
            elif k[0][0] == "[":
                if re.match(k, stage_code):
                    if match_key is not None:
                        msg = 'There is an overlap in stage_modifications'\
                              f' for stage {sequence_number}. Taking the '\
                              f'first applicable pattern = "{match_key}"'
                        warnings.warn(msg)
                        logger.warning(msg)
                        break  # Only use first match, to avoid conflicts
                    range_dict = v
                    match_key = k

        # Gather all modifications in a single dict
        # Do this in order: particular mods have priority over range
        # specific which has priority over default
        for k, v in range_dict.items():
            if k not in modif:
                modif[k] = v

        for k, v in default_dict.items():
            if k not in modif:
                modif[k] = v

        return modif




"""
Filter class and subclasses
"""
# Standard library modules
import math as m
import logging

# Non-standard modules
from .filter_template import FilterTemplate
from ...helpers import str_indent

logger = logging.getLogger("obsinfo")


class PolesZeros(FilterTemplate):
    """
    PolesZeros filter

    Attributes:
        transfer_function_type (str): one of  'LAPLACE (RADIANS/SECOND)',
            'LAPLACE (HERTZ)','DIGITAL (Z-TRANSFORM)'
        poles (list of complex numbers)
        zeros (list of complex numbers)
        normalization_frequency (float)
        normalization_factor (float)
    """

    def __init__(self, attr_dict, stage_id=-1):
        """
        Constructor

        Args:
            attr_dict (dict or :class:`.ObsMetadata`): filter attributes
            stage_id (int): id of corresponding stage. Used for reporting only
        """
        logger.debug(f'in {self.__class__.__name__}.__init__()')
        super().__init__(attr_dict, stage_id)
        self.transfer_function_type = attr_dict.pop(
            'transfer_function_type', 'LAPLACE (RADIANS/SECOND)')
        self.poles = [complex(x.replace(" ", ""))
                      for x in attr_dict.pop('poles', [])]
        self.zeros = [complex(x.replace(" ", ""))
                      for x in attr_dict.pop('zeros', [])]
        # sensitivity_gain_frequency is handed down by Filter superclass,
        # in case the normalization frequency is not specified
        self.normalization_frequency = attr_dict.pop('sensitivity_gain_frequency', 1.)
        if 'normalization_frequency' in attr_dict:  # Overrides sensitivity_gain_frequency
            self.normalization_frequency = attr_dict.pop('normalization_frequency')
        if self.normalization_frequency == 0:
            self.normalization_frequency = 1
        self.normalization_factor = attr_dict.pop('normalization_factor', None)
        self._validate_empty_attributes_dict(attr_dict)

        # poles and zeros should be lists of complex numbers
        if self.transfer_function_type not in ['LAPLACE (RADIANS/SECOND)',
                                               'LAPLACE (HERTZ)',
                                               'DIGITAL (Z-TRANSFORM)']:
            msg = (f'Illegal transfer_function_type in PolesZeros: '
                   f'"{self.transfer_function_type}" in stage #{stage_id}')
            logger.error(msg)
            raise TypeError(msg)

        # Set/calculate normalization factor
        if not (self.normalization_frequency and self.normalization_factor):
            self.normalization_factor = self.calc_normalization_factor(stage_id)

    def __str__(self, indent=0, n_subclasses=0):
        """
        Args:
            indent (int): number of extra characters to indent lines by
            n_subclasses (int): number of levels of subclass to print out
        """
        if n_subclasses < 0:
            return f'{type(self)}'
        s = super().__str__() + '\n'
        s += f'    {len(self.poles)} poles, {len(self.zeros)} zeros\n'
        s += f'    transfer_function_type: {self.transfer_function_type}\n'
        s += f'    normalization_frequency: {self.normalization_frequency:g}\n'
        s += f'    normalization_factor: {self.normalization_factor:g}'
        return str_indent(s, indent)

    def calc_normalization_factor(self, stage_id=-1, debug=False):
        """
        Calculate the normalization factor for a given set of poles-zeros

        The norm factor A0 is calculated such that

        .. parsed-literal::
                                  sequence_product_over_n(s - zero_n)
                       A0 * abs(—--------------------------------------) == 1
                                  sequence_product_over_m(s - pole_m)

            for s_f = i*2pi*f if the transfer function is Laplace in radians
                      i*f if the transfer function is Laplace in Hertz

        There is no calculation for the digital z-transform

        Returns:
            normalization factor as a float or None if not Laplace
        """
        if not self.normalization_frequency:
            msg = (f'No normalization frequency in {self.__class__.__name__} at stage "{stage_id}"')
            logger.error(msg)
            raise ValueError(msg)

        A0 = 1.0 + (1j * 0.0)
        if self.transfer_function_type == "LAPLACE (HERTZ)":
            s = 1j * self.normalization_frequency
        elif self.transfer_function_type == "LAPLACE (RADIANS/SECOND)":
            s = 1j * 2 * m.pi * self.normalization_frequency
        else:
            logger.warning("Don't know how to calculate normalization factor"
                           "for z-transform poles and zeros!")
            return None

        for p in self.poles:
            A0 *= (s - p)
        for z in self.zeros:
            A0 /= (s - z)

        logger.debug(f"poles={self.poles}, zeros={self.zeros}, s={s}, A0={A0}")

        A0 = abs(A0)
        return A0



"""
Filter classes:
- Coefficients
- FIR
- PolesZeros
- ResponseList
- Polynomial (never tested)
- ADConversion (subclass of PolesZeros)
- Analog (subclass of PolesZeros)
- Digital (subclass of Coefficients)
"""
from .filter import Filter
from .coefficients import Coefficients
from .FIR import FIR
from .poles_zeros import PolesZeros
from .response_list import ResponseList
from .polynomial import Polynomial
from .AD_conversion import ADConversion
from .analog import Analog
from .digital import Digital



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Application to print obsinfo information files.
"""
import os
import warnings
from pathlib import Path
from argparse import ArgumentParser

# obsinfo modules
from ..obsmetadata import ObsMetadata
from ..instrumentation import (Instrumentation, Stage, Filter,
                                     Preamplifier, Sensor, Datalogger)
from ..helpers import Location, Locations, Person, init_logging
from ..subnetwork import (Subnetwork, Network, Operator)
from ..misc.discoveryfiles import Datapath

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
verbose = False

logger = init_logging("print", console_level='WARNING')


def print_single_file(info_file, n_sublevels=0, configs=False, verbose=True):
    """
    Print a single obsinfo file.

    Args:
        info_file (str): info file to validate.
        n_sublevels (int): number of sublevels to print_directory
        configs (boolean): print configurations instead of object
    """
    filetype = ObsMetadata.get_information_file_type(info_file)
    try:
        dp = Datapath()
        attributes_dict = ObsMetadata().read_info_file(info_file, dp, quiet=True)
    except Exception as e:
        print(str(e))
        return False
    if configs is not True:
        print(f'{Path(info_file).name}: ', end='', flush=True)
        no_class_msg = f"Can't print {filetype} files (no associated class)"
        if filetype == 'subnetwork':
            obj = Subnetwork(attributes_dict[filetype])
        elif filetype == 'network':
            obj = Network(attributes_dict[filetype])
        elif filetype == 'instrumentation_base':
            location_code='00'
            locations = Locations(
                [Location({'position': {'lat': 0, 'lon': 0, 'elev': 0},
                           'base': {'uncertainties.m': {'lat': 0, 'lon': 0, 'elev':0},
                                    'depth.m': 0,
                                    'geology': 'seafloor',
                                    'vault': 'seafloor'},
                           'code': '00'})
                ])
            obj = Instrumentation({'base': attributes_dict[filetype]},
                                  locations, location_code,
                                  '2000-01-01', '3000-01-01')
        elif filetype == 'datalogger_base':
            obj = Datalogger({'base': attributes_dict[filetype]})
        elif filetype == 'sensor_base':
            obj = Sensor({'base': attributes_dict[filetype]})
        elif filetype == 'preamplifier_base':
            obj = Preamplifier({'base': attributes_dict[filetype]})
        elif filetype == 'location_base':
            obj = Location({'base': attributes_dict[filetype],
                             'position': {'lat': 0, 'lon': 0, 'elev': 0}})
        elif filetype == 'stage_base':
            obj = Stage({'base': attributes_dict[filetype]})
        elif filetype == 'network':
            print(no_class_msg)
            return
        elif filetype == 'operator':
            obj = Operator(attributes_dict[filetype])
        elif filetype == 'filter':
            obj = Filter.construct(attributes_dict[filetype], 1, 'print')
        elif filetype == 'person':
            obj = Person(attributes_dict[filetype])
        elif filetype == 'timing_base':
            print(no_class_msg)
            return
        else:
            print(f"Unknown information file type: '{filetype}")
            return
        print("\n    " + obj.__str__(indent=4, n_subclasses=n_sublevels))
    else:
        if filetype in ('instrumentation_base', 'datalogger_base',
                        'sensor_base' 'preamplifier_base',
                        'location_base', 'stage_base', 'timing_base'):
            configs = attributes_dict[filetype].get('configurations', {})
            keys = list(configs.keys())
            if verbose or len(keys) > 0:
                print(f'{Path(info_file).name}: ', end='', flush=True)
                print(f"{list(configs.keys())}")
        else:
            if verbose:
                print(f'{Path(info_file).name}: ', end='', flush=True)
                print(f"file type: '{filetype}' has no configurations")
            return


def print_directory(info_dir, n_sublevels=1, configs=False,
                    drilldown=False, verbose=False):
    """
    Validate all information files in a directory.

    Args:
        info_dir (:class:`Path`): directory where files reside
        drilldown (bool): drill down through subdirectories
        n_sublevels (int): number of sublevels to print_directory
        configs (boolean): print configurations instead of object
    Returns:
        false if any file failed
    """
    suffixes = ['.yaml', '.json']
    if not info_dir.is_dir():
        raise ValueError(f'"{info_dir}" is not a directory!')
    if drilldown is True:
        print(f'Printing files in and below directory {info_dir}')
        files = info_dir.glob('**/*.*')
    else:
        print(f'Validating files in directory {info_dir}')
        files = info_dir.glob('*.*')
    print(f'Will skip files whose suffix is not in "{suffixes}"')
    for file in files:
        if file.suffix.lower() not in suffixes:
            print(f'Skipping {str(file.relative_to(info_dir))}')
            continue
        print_single_file(str(file), n_sublevels=n_sublevels,
                          configs=configs, verbose=verbose)


def main():
    """
    Entry point for obsinfo-print. Print an information file according to its
    type, and levels below up to the specified level.

     Captures all exceptions

    """

    args = retrieve_arguments()

    if args.input.is_file():
        print_single_file(str(args.input), args.n_levels, args.configs,
                          args.verbose)
    elif args.input.is_dir():
        print_directory(args.input, args.n_levels, args.configs, args.drilldown,
                        args.verbose)


def retrieve_arguments():
    """
    Retrieve arguments from command line.

    Setup several status variables and get information file name

    Returns:
        (dict): all status variables and information file name.
    """

    # Parse the arguments
    parser = ArgumentParser(prog="obsinfo-print")

    # flags
    parser.add_argument("-n", "--n_levels", type=int, default=1,
                        help="Prints up to N levels")
    parser.add_argument("-d", "--debug", action='store_true', default=False,
                        help="Print traceback for exceptions")
    parser.add_argument("--drilldown", action='store_true', default=False,
                        help="Drill down through all subdirectories (if a "
                        "directory was specified)")
    parser.add_argument("--configs", action='store_true', default=False,
                        help="Print configurations instead of object")
    parser.add_argument("--verbose", action='store_true', default=False,
                        help="Be verbose")
    # positional arguments
    parser.add_argument("input", type=str,
                        help="Information file or directory to be validated. "
                             "If a directory, tests all files in the "
                             "directory.  If 'DATAPATH', will test all files "
                             "in the DATAPATH (sets --drilldown and "
                             "--n_sublevels=0)")

    args = parser.parse_args()

    if args.input == 'DATAPATH':
        args.input = Datapath().datapath_list[0]
        args.drilldown = True
        print('Validating first DATAPATH dir')
    args.input = Path(args.input)
    if not args.input.is_absolute():
        args.input = Path(os.getcwd()).joinpath(args.input).resolve()

    return args


if __name__ == '__main__':
    main()



"""
Comments Class
"""
# Standard library modules
import warnings
import logging
import json

from obspy.core.inventory.util import Comment as obspy_comment

from .functions import str_indent
from .person import Persons
from .oi_date import OIDate
from .obsinfo_class_list import ObsinfoClassList

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Comments(ObsinfoClassList):
    """
    A list of Comment objects
    """
    def __init__(self, attributes_list):
        """
            Args:
                attributes_list: (list of dict or str): list
                    of comments read from YAML/JSON file
        """
        if attributes_list is None:
            super().__init__([], Comment)
        else:
            super().__init__([Comment(x) for x in attributes_list], Comment)

    @classmethod
    def from_extras(cls, extras_dict):
        """
        Create Comments from "extras" (freeform dict object)
        """
        if extras_dict is None:
            return cls(None)
        else:
            return cls(['Extra attributes: ' + json.dumps(extras_dict)])


class Comment(object):
    def __init__(self, attributes_dict):
        if isinstance(attributes_dict, str):
            self.value = attributes_dict
            self.authors = Persons(None)
            self.begin_effective_time = OIDate(None)
            self.end_effective_time = OIDate(None)
            self.id = None
            self.subject = None
        else:
            self.value = attributes_dict.pop('value')
            self.authors = Persons(attributes_dict.pop('authors', None))
            self.begin_effective_time = OIDate(attributes_dict.pop('begin_effective_time', None))
            self.end_effective_time = OIDate(attributes_dict.pop('end_effective_time', None))
            self.id = attributes_dict.pop('id', None)
            self.subject = attributes_dict.pop('subject', None)

    def __str__(self, indent=0, n_subclasses=0):
        if n_subclasses < 0:
            return(f'{self.__class__.__name__}: {self.value}')
        s = 'Comment:\n'
        s += f'    value: {self.value}'
        if len(self.authors) > 0:
            s += f'\n    authors: {self.authors.__str__(indent=4, n_subclasses=n_subclasses-1)}'
        if self.begin_effective_time.to_obspy() is not None:
            s += f'\n    begin_effective_time: {self.begin_effective_time}'
        if self.end_effective_time.to_obspy() is not None:
            s += f'\n    end_effective_time: {self.end_effective_time}'
        if self.id is not None:
            s += f'\n    id: {self.id}'
        if self.subject is not None:
            s += f'\n    subject: {self.subject}'
        return str_indent(s, indent)

    def to_obspy(self):
        return obspy_comment(
            value=self.value,
            begin_effective_time=self.begin_effective_time.to_obspy(),
            end_effective_time=self.end_effective_time.to_obspy(),
            authors=self.authors.to_obspy(),
            subject=self.subject
        )



"""
Generate scripts needed to go from basic miniSEED to data center ready
"""
# import obsinfo
from obsinfo.subnetwork import Subnetwork
import os.path
import sys
from obspy.core import UTCDateTime

SEPARATOR_LINE = "\n# " + 60 * "=" + "\n"


def process_script(
    station,
    station_dir,
    distrib_dir="/opt/sdpchain",
    input_dir="miniseed_basic",
    corrected_dir="miniseed_corrected",
    extra_commands=None,
    include_header=True,
    SDS_uncorr_dir="SDS_uncorrected",
    SDS_corr_dir="SDS_corrected",
    SDS_combined_dir="SDS_combined",
):
    """
    Writes OBS data processing script using SDPCHAIN software

    Args:
        station:          an obsinfo.station object
        station_dir:      base directory for the station data
        input_dir:        directory beneath station_dir for input (basic)
                          miniseed data ['miniseed_basic']
        corrected_dir:    directory beneath station_dir for output (corrected)
                          miniseed data ['miniseed_corrected']
        SDS_corr_dir:     directory beneath station_dir in which to write
                          SDS structure of corrected data (ideally
                          ../SOMETHING if ms2sds could write all to the
                          same directory)
        SDS_uncorr_dir:   directory beneath station_dir in which to write
                          SDS structure of uncorrected data (ideally
                          ../SOMETHING if ms2sds could write all to the
                          same directory)
        include_header:   whether or not to include the bash script header
                          ('#!/bin/bash') at the top of the script [True]
        distrib_dir:      Base directory of sdpchain distribution ['/opt/sdpchain']

    The sequence of commands is:
        1. optional proprietary format steps (proprietary format -> basic miniseed, separate)
        2. optional extra_steps (any cleanup needed for the basic
           miniseed data, should either overwrite the existing data or
           remove the original files so that subsequent steps only see the
           cleaned data)
        3. ms2sds on basic miniseed data
        4. leap-second corrections, if necessary
        5. msdrift (creates drift-corrected miniseed)
    """
    leap_corr_dir = "miniseed_leap_corrected"

    s = ""
    if include_header:
        s += __header(station.code)
    s += __setup_variables(distrib_dir, station_dir)
    if extra_commands:
        s += __extra_command_steps(extra_commands)
    s += __ms2sds_script(station, input_dir, SDS_uncorr_dir)
    t = __leap_second_script(
        station.clock_corrections.get("leapseconds", None), input_dir, leap_corr_dir
    )
    if t:
        s += t
        input_dir = leap_corr_dir
    s += __msdrift_script(input_dir, corrected_dir, station.clock_corrections)
    s += __force_quality_script(corrected_dir, "Q")
    s += __ms2sds_script(station, corrected_dir, SDS_corr_dir)
    s += __combine_sds_script(station, SDS_corr_dir, SDS_uncorr_dir, SDS_combined_dir)

    return s


############################################################################
def __header(station_name):

    s = "#!/bin/bash\n"
    s += SEPARATOR_LINE + f'echo "Working on station {station_name}"' + SEPARATOR_LINE

    return s


def __setup_variables(distrib_dir, station_dir):

    s = SEPARATOR_LINE + "# SDPCHAIN STEPS" + SEPARATOR_LINE
    s += "#  - Set up paths\n"
    s += f"STATION_DIR={station_dir}\n"
    s += f"MSDRIFT_EXEC={os.path.join(distrib_dir,'bin','msdrift')}\n"
    s += f"MSDRIFT_CONFIG={os.path.join(distrib_dir,'config','msdrift.properties')}\n"
    s += f"MS2SDS_EXEC={os.path.join(distrib_dir,'bin','ms2sds')}\n"
    s += f"MS2SDS_CONFIG={os.path.join(distrib_dir,'config','ms2sds.properties')}\n"
    s += f"SDPPROCESS_EXEC={os.path.join(distrib_dir,'bin','sdp-process')}\n"
    s += f"MSMOD_EXEC={os.path.join('/opt/iris','bin','msmod')}\n"
    s += "\n"

    return s


def __extra_command_steps(extra_commands):
    """
    Write extra command lines, embedded in sdp-process

    Input:
        extra_commands: list of strings containing extra commands
    """
    s = SEPARATOR_LINE
    s = s + "# - EXTRA COMMANDS\n"
    if not isinstance(extra_commands, list):
        raise ValueError("extra_commands is not a list")
    for cmd_line in extra_commands:
        s = s + '$SDPPROCESS_EXEC --cmd="{cmd_line}"\n'
    return s


def __ms2sds_script(station, in_path, out_path):

    """
    Writes the ms2sds lines of the script
    """
    sta = station.code
    net = station.network_code

    s = f'echo "{"-"*60}"\n'
    s += 'echo "Running MS2SDS: MAKE SDS ARCHIVE"\n'
    s += f'echo "{"-"*60}"\n'

    s += f'in_dir="{in_path}"\n'
    s += f'out_dir="{out_path}"\n'

    s += "# - Create output directory\n"
    s += "mkdir $STATION_DIR/$out_dir\n"

    s += "# - Collect input filenames\n"
    s += "command cd $STATION_DIR/$in_dir\n"
    s += "mseedfiles=$(ls *.mseed)\n"
    s += "command cd -\n"
    s += 'echo "mseedfiles=" $mseedfiles\n'

    s += "# - Run executable\n"
    s += "$MS2SDS_EXEC $mseedfiles -d $STATION_DIR -i $in_dir -o $out_dir "
    s += f'--network "{net}" --station "{sta}" -a SDS -p $MS2SDS_CONFIG\n'
    s += "\n"

    return s


############################################################################
def __leap_second_script(leapseconds, in_dir, out_dir):
    """
    Create leap-second correction text

    Inputs:
        leapseconds: list of dictionaries from network information file
    """
    if not leapseconds:
        return ""

    s = f'echo "{"-"*60}"\n'
    s += 'echo "LEAPSECOND CORRECTIONS"\n'
    s += f'echo "{"-"*60}"\n'

    s += f"in_dir={in_dir}\n"
    s += f"out_dir={out_dir}\n"

    s += "# - Create output directory\n"
    s += "mkdir $STATION_DIR/$out_dir\n"

    s += "# - Copy files to output directory\n"
    s += "cp $STATION_DIR/$in_dir/*.mseed $STATION_DIR/$out_dir\n"

    for leapsecond in leapseconds:
        if leapsecond["corrected_in_basic_miniseed"]:
            s += "# LEAP SECOND AT {} ALREADY CORRECTED IN BASIC MINISEED, DOING NOTHING\n".format(
                leapsecond["time"]
            )
            return s
        temp = leapsecond["time"].split("T")
        d = UTCDateTime(temp[0])
        leap_time = d.strftime("%Y,%j,") + temp[1].rstrip("Z")
        s += 'echo ""\n'
        s += f'echo "{"="*60}"\n'
        s += 'echo "Running LEAPSECOND correction"\n'
        s += f'echo "{"-"*60}"\n'
        if leapsecond["type"] == "+":
            s += '$SDPPROCESS_EXEC -d $STATION_DIR -c="Shifting one second BACKWARDS after positive leapsecond" '
            s += f' --cmd="$MSMOD_EXEC --timeshift -1 -ts {leap_time} -s -i $STATION_DIR/$out_dir/*.mseed"\n'
            s += '$SDPPROCESS_EXEC -d $STATION_DIR -c="Marking the record containing the positive leapsecond" '
            s += f' --cmd="$MSMOD_EXEC --actflags 4,1 -tsc {leap_time} -tec {leap_time} -s -i $STATION_DIR/$out_dir/*.mseed"\n'
        elif leapsecond["type"] == "-":
            s += '$SDPPROCESS_EXEC -d $STATION_DIR -c="Shifting one second FORWARDS after negative leapsecond" '
            s += f' --cmd="$MSMOD_EXEC --timeshift +1 -ts {leap_time} -s -i $STATION_DIR/$out_dir/*.mseed"\n'
            s += '$SDPPROCESS_EXEC -d $STATION_DIR -c="Marking the record containing the negative leapsecond" '
            s += f' --cmd="$MSMOD_EXEC --actflags 5,1 -tsc {leap_time} -tec {leap_time} -s -i $STATION_DIR/$out_dir/*.mseed"\n'
        else:
            s += 'ERROR: leapsecond type "{}" is neither "+" nor "-"\n'.format(
                leapsecond["type"]
            )
            sys.exit(2)
    return s


def __msdrift_script(in_path, out_path, clock_corrs):
    """
    Write msdrift lines of the script

    Inputs:
        in_path
        out_path
        clock_corrs
    """
    s = f'echo "{"-"*60}"\n'
    s += 'echo "Running MSDRIFT: CORRECT LINEAR CLOCK DRIFT"\n'
    s += f'echo "{"-"*60}"\n'

    s += f'in_dir="{in_path}"\n'
    s += f'out_dir="{out_path}"\n'

    s += "# - Create output directory\n"
    s += "mkdir $STATION_DIR/$out_dir\n"

    s += "# - Collect input filenames\n"
    s += "command cd $STATION_DIR/$in_dir\n"
    s += "mseedfiles=$(ls *.mseed)\n"
    s += "command cd -\n"
    s += 'echo "mseedfiles=" $mseedfiles\n'

    if "linear_drift" in clock_corrs:
        lin_corr = clock_corrs["linear_drift"]
        s += "# - Run executable\n"
        s += f'START_REFR="{str(lin_corr["start_sync_reference"]).rstrip("Z")}"\n'
        s += f'START_INST="{str(lin_corr["start_sync_instrument"]).rstrip("Z")}"\n'
        s += f'END_REFR="{str(lin_corr["end_sync_reference"]).rstrip("Z")}"\n'
        s += f'END_INST="{str(lin_corr["end_sync_instrument"]).rstrip("Z")}"\n'
        s += "$MSDRIFT_EXEC $mseedfiles -d $STATION_DIR -i $in_dir -o $out_dir "
        s += '-m "%E.%S.00.%C.%Y.%D.%T.mseed:%E.%S.00.%C.%Y.%D.%T_driftcorr.mseed" '
        s += '-s "$START_REFR;$START_INST" '
        s += '-e "$END_REFR;$END_INST" '
        s += "-p $MSDRIFT_CONFIG\n"
        s += "\n"
    else:
        while lin_corr in clock_corrs["linear_drifts"]:
            s += SEPARATOR_LINE
            s += "ERROR, CANT YET APPLY MULTIPLE TIME CORRECTIONS (SHOULD CHANGE\n"
            s += "MSDRIFT TO ONLY WRITE GIVEN TIME RANGE AND BE ABLE TO APPEND TO EXISTING FILE?)\n"
            s += SEPARATOR_LINE
    return s


def __force_quality_script(in_path, quality="Q"):
    """
    Force data quality to Q

    Inputs:
        in_path: directory where mseed files are found
        quality: new quality code
    """
    s = f'echo "{"-"*60}"\n'
    s += 'echo "Forcing data quality to Q"\n'
    s += f'echo "{"-"*60}"\n'
    s += f'$SDPPROCESS_EXEC -d $STATION_DIR -c="Forcing data quality to Q" --cmd="$MSMOD_EXEC --quality Q -i {in_path}/*.mseed"\n'
    s += "\n"

    return s


def __combine_sds_script(station, SDS_corrected_dir, SDS_uncorrected_dir,
                         SDS_combined_dir):
    """
    Combine corrected and uncorrected SDS directories/files

    Corrected must be before uncorrected so that Arclink will extract it by
    default
    Uses the corrected directory as the template: will not process uncorrected
    files that don't have corresponding corrected files

    Creates no process-steps because there are so many individual calls to
    msmod (should we write an sdp-sds_combine script?)

    Inputs:
        station:             station object
        SDS_corrected_dir:   name of SDS directory containing corrected data
        SDS_uncorrected_dir: name of SDS directory containing uncorrected data
        SDS_combined_dir:    name of SDS directory to put combined data into
    """
    s = f'echo "{"-"*60}"\n'
    s += 'echo "Combining corrected and uncorrected SDS datafiles"\n'
    s += f'echo "{"-"*60}"\n'
    s += f"mkdir -p $STATION_DIR/{SDS_combined_dir}\n"
    s += f"command cd $STATION_DIR/{SDS_corrected_dir}\n"
    s += "years=????\n"
    s += "command cd -\n"
    s += "for y in $years ; do\n"
    s += f"    for d in $STATION_DIR/{SDS_corrected_dir}/SDS/$y/{station.network_code}/{station.code}/*.D ; do\n"
    s += "        d_sub=${{d#$STATION_DIR/{}/SDS/}}\n".format(SDS_corrected_dir)
    s += f"        if [[ ! -d $STATION_DIR/{SDS_combined_dir}/$d_sub ]] ; then\n"
    s += '            echo "Creating and filling combined SDS subdirectory $d_sub"\n'
    s += f'            eval "mkdir -p $STATION_DIR/{SDS_combined_dir}/$d_sub"\n'
    s += "            for f in $d/*; do\n"
    s += "                f_sub=${{f#$STATION_DIR/{}/SDS/}}\n".format(SDS_corrected_dir)
    s += f"                cat $STATION_DIR/{SDS_corrected_dir}/SDS/$f_sub "
    s += f"$STATION_DIR/{SDS_uncorrected_dir}/SDS/$f_sub > "
    s += f"$STATION_DIR/{SDS_combined_dir}/$f_sub\n"
    s += "            done\n"
    s += "        else\n"
    s += '            echo "Combined SDS subdirectory $d_sub already exists, skipping..."\n'
    s += "        fi\n"
    s += "    done\n"
    s += "done\n"
    s += "\n"
    return s


def _console_script(argv=None):
    """
    Console-level processing script

    requires SDPCHAIN programs msdrift and ms2sds, and IRIS program msmod

    Currently uses 'SDS_corrected' and 'SDS_uncorrected' as default SDS
    directories.
    Would be better to use ../SDS_[un]corrected so that all data are together,
    but ms2sds is not yet capable of putting data in an existing SDS directory

    """
    from argparse import ArgumentParser

    parser = ArgumentParser(prog="obsinfo-make_process_scripts_SDPCHAIN",
                            description=__doc__)
    parser.add_argument("subnetwork_file", help="Subetwork information file")
    parser.add_argument("station_data_path",
                        help="Base path containing station data")
    parser.add_argument("distrib_dir", default="/opt/sdpchain",
                        help="Path to SDPCHAIN software")
    parser.add_argument("-i", "--input_dir", default="miniseed_basic",
                        help="subdirectory of station_data_path/{STATION}/ "
                             "containing input *.mseed files")
    parser.add_argument("-o", "--corrected_dir", default="miniseed_corrected",
                        help="subdirectory of station_data_path/{STATION}/ "
                             "to put corrected *.mseed files")
    parser.add_argument("--SDS_corr_dir", default="SDS_corrected",
                        help="subdirectory of station_data_path/{STATION}/ "
                             "for SDS structure of corrected data")
    parser.add_argument("--SDS_uncorr_dir", default="SDS_uncorrected",
                        help="subdirectory of station_data_path/{STATION}/ "
                             "for SDS structure of uncorrected data")
    parser.add_argument("--suffix", default="_SDPCHAIN",
                        help="suffix for script filename")
    parser.add_argument("--append", action="store_true",
                        help="append to existing script file")
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="increase output verbosity")
    parser.add_argument("--no_header", action="store_true",
                        help="do not include file header")
    parser.add_argument("-q", "--quiet", action="store_true",
                        help="run silently")
    args = parser.parse_args()

    # READ IN NETWORK INFORMATION
    if not args.quiet:
        print("Creating SDPCHAIN process scripts, ", end="")
    subnetwork = Subnetwork(args.subnetwork_file)
    if not args.quiet:
        print(f"subnetwork {subnetwork.network.code}, stations ", end="")
        if args.verbose:
            print("")

    first_time = True
    for name, station in subnetwork.stations.items():
        if not args.quiet:
            if args.verbose:
                print(f"\t{name}", end="")
            else:
                if first_time:
                    print(f"{name}", end="")
                else:
                    print(f", {name}", end="")
        station_dir = os.path.join(args.station_data_path, name)
        script = process_script(
            station,
            station_dir,
            distrib_dir=args.distrib_dir,
            input_dir=args.input_dir,
            corrected_dir=args.corrected_dir,
            SDS_uncorr_dir=args.SDS_uncorr_dir,
            SDS_corr_dir=args.SDS_corr_dir,
            include_header=not args.no_header,
        )
        fname = "process_" + name + args.suffix + ".sh"
        if args.verbose:
            print(f" ... writing file {fname}", flush=True)
        if args.append:
            write_mode = "a"
        else:
            write_mode = "w"
        with open(fname, write_mode) as f:
            f.write(script)
            f.close()
        first_time = False
    if not args.verbose and not args.quiet:
        print("")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Application to test and print obsinfo information files.

Includes class TestDatapath and entry points.

   # TestObsinfo tests information files content (should only tests
     DATAPATH-accessed functions)

 Uses $HOME/.obsinforc to determine the information file base, so if you
 want to test the example files, extract them using obsinfo-setup -d or
 put the path to the example files in $HOME/.obsinforc

 Much the SAME functions and methods are called whether printing or testing,
 as there is little difference between both functionalities. Entry points vary
 for each functionality and the distinction is achieved by naming
 the executable with two different names (i.e. a named link):


    * obsinfo-print has entry point print_obs.

    * obsinfo-test has entry point run_suite_info_files. It is meant to be
      used by developers. In this vein, contrary to obsinfo-validate,
      obsinfo-print calls private methods which are not callable from the
      command line but which can be called from the interpreter.

    * JsonRefTest is meant to be called from the interpreter. It has no entry
      point.


 There are two types of testing functionalities.

    a) If the file name includes "test--attributes" the output of the
       corresponding obsinfo test function will be checked against data
       contained in this class.

    b) If the file name is "normal", it will simply run through to make sure
       there are no errors

 Testing for type (a) uses data from
    obsinfo/tests/data/Information_Files/responses/_filters.
 Testing for type (b) uses data from obsinfo/_examples/Information_Files

 WARNING: many tests are critically dependent on file hierarchy, including
 names. Do not change names
 in tests or _examples hierarchy, or else change the names here.
 Also, the following methods use four specific file names:

     * test_all_stage_types()
     * test_sensor()
     * test_preamplifier()
     * test_datalogger()
     * test_station()

 All current examples of sensors except NANOMETRICS_T240_SINGLESIDED have no
 configuration key and no default.
 Messages to this effect are to be expected.
"""

# from __future__ import (absolute_import, division, print_function,
#                         unicode_literals)
# from future.builtins import *  # NOQA @UnusedWildImport
# 
# import os
import warnings

# import sys
from pathlib import Path, PurePath
import unittest
# import inspect
# import difflib
import re
# import glob
# from json.decoder import JSONDecodeError
# from argparse import ArgumentParser

import logging
# from logging.handlers import RotatingFileHandler

# Third party imports
from obspy.core.utcdatetime import UTCDateTime
from obspy.core.inventory.util import Site

# obsinfo modules
from ..obsmetadata import (ObsMetadata)
from ..instrumentation import (Instrumentation, InstrumentComponent,
                                     Stages, Stage, Filter)
from ..helpers import Location
from ..subnetwork import (Station, Subnetwork)
from ..instrumentation.filter import (PolesZeros, FIR, Coefficients, ResponseList)
from ..misc.printobs import (PrintObs)
from ..misc.discoveryfiles import Datapath

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
verbose = False


class TestObsinfo(unittest.TestCase):
    """
    Test suite and print methods for obsinfo operations.

    Attributes:
        infofiles_path (str): path to datafiles to be tested
        level (str): level to be printed
        test (boolean): determines if this is test mode
        print_output (boolean): determines if this is print mode.
            Both can coexist.
    """

    def setUp(self, test=True, print_output=False, level=None):
        """
        Set up default values and paths

        Args:
            test (bool): Invoke class methods in test mode, i.e. invoke
                assertions as well as obsinfo object creation
            print_output (bool): Invoke class methods in print mode, no
                                 assertions.
            level (str): In print mode, determine up to which level
                         information will be printed
        """
        self.infofiles_path = Datapath()

        self.level = level
        self.test = test
        self.print_output = print_output

    def test_test_filters(self):
        """
        Test all information files in test/data/Information_Files/_filters"

        If you wish to test individual files, use test_filter(file) with file an absolute or
        relative file name.
        """
        warnings.warn('Not testing local information files')
        return
        globs = Path(__file__).parent.joinpath("data", "Information_Files", "filters").glob("*/*.yaml")
        for file in [str(x.resolve()) for x in globs]:
            self._test_filter(file)

    def _test_filter(self, info_file):
        """
        Test and/or print a filter file.

        All are actual examples except the info files called "test---attributes.
        In this special cases there will also be a comparison against a dict
        of expected results.
        This comparison occurs for all four main types of filters.

        Args:
            info_file (str): Filename to test or print
        """
        test_expected_result = {
            'PolesZeros' :
                {
                    "type": "PolesZeros",
                    "transfer_function_type": "LAPLACE (RADIANS/SECOND)",
                    "zeros": [(0.0+0.0j)],
                    "poles": [(0.546+0.191j), (4.00004e4+0.000j),],
                    "normalization_frequency" : 1.,
                    "normalization_factor" : 42833.458122775904,
                    "delay.seconds": 0,
                },
            'FIR' :
                {
                    "type": "FIR",
                    "symmetry": "ODD",
                    "coefficient_divisor": 8388608,
                    "coefficients": [-10944, 0, 103807, 0, -507903, 0, 2512192, 4194304,],
                    "delay.samples": 7,
                },
            'Coefficients' :
                {
                    "type" : "Coefficients",
                    "transfer_function_type" : "DIGITAL",
                    "numerator_coefficients" :   [1, 0.1, -0.3, 0.6],
                    "denominator_coefficients" : [-0.2, 0.8, 0.4, -0.3],
                    "delay.samples": 0,
                },
            'ResponseList' :
                {
                    "type" : "ResponseList",
                    "delay.samples": 0,
                    "response_list" : [[ 0.050, 0.56, 0.0],
                                       [ 0.075, 0.73, 0.0],
                                       [ 1, 1, 0.0],
                                       [10, 0.97, -179],
                                       [100, 0.96, 179],
                                       [1000, 0.96, 179],
                                       [5000, 0.82, 143],
                                       [7500, 0.69, 129]],
                },
            'AD_CONVERSION' :
                {
                    "type" : "AD_CONVERSION",
                    "input_full_scale" : 5,
                    "output_full_scale" : 4294967292,
                    "transfer_function_type" : "DIGITAL",
                    "numerator_coefficients" :   [1.0],
                    "denominator_coefficients" : [],
                    "delay.samples": 0,
                },
            'ANALOG' :
                {
                    "type" : "ANALOG",
                    "transfer_function_type" : "LAPLACE (RADIANS/SECOND)",
                    "zeros": [],
                    "poles": [],
                    "normalization_frequency" : 0.,
                    "normalization_factor" : 1.0,
                    "delay.seconds": 0,
                },
            'DIGITAL' :
                {
                    "type" : "DIGITAL",
                    "transfer_function_type" : "DIGITAL",
                    "numerator_coefficients" :   [1.0],
                    "denominator_coefficients" : [],
                    "delay.samples": 0,
                },
        }

        read_stream = ObsMetadata.read_info_file(info_file, self.infofiles_path)
        obj = Filter.construct(ObsMetadata(read_stream['filter']), "")

        if verbose:
           print(f'Processing filter file:"{info_file}"')

        if self.test:
            # self.assertTrue(isinstance(obj, obj.type),
            #                 f'Object {info_file} is not a {obj.type} filter')

            # compare with expected result
            if re.match("test---attributes", str(info_file)):
                self._filter_compare(info_file, obj, test_expected_result)

        if verbose:
            print(f"Filter test for: {info_file}: PASSED")

        if self.print_output:
           print(obj)

    def _filter_compare(self, info_file, filter, expected_result):
        """
        Test a created filter object against an expected result

        :param: info_file - Filename to test or print
        :type info_file: str
        :param: filter - type of filter
        :type filter: str
        :param: expected_result - dictionary containing attributes to test against
        :type expected_result: dict
        """
        ftype = filter.type
        read_dict = vars(filter)

        # Remove notes and extras
        if read_dict.get('notes', None) == []:
            read_dict.pop('notes')
        if read_dict.get('extras', None) is None:
            read_dict.pop('extras')

        self.assertEqual(read_dict,
                         expected_result.get(ftype, None),
                         f" File: '{info_file}'. Computed result: {read_dict} "
                         "and expected result: "
                         f"{expected_result.get(ftype, None)} are different")

    def _test_PZ_conditionals(self):
        """
        Test all the conditionals in the PZ filter. In particular, this tests the function
        to calculate the normalization factor.
        """
        obj = Filter.construct(
            ObsMetadata({'type': 'PolesZeros',
                         'transfer_function_type' : 'LAPLACE (RADIANS/SECOND)',
                         'zeros': ['0.3+0.2j'],
                         'poles': ['0.546+0.191j', '4.40e4+0.000j'],
                         'normalization_frequency' : 1.,
                         'normalization_factor' : None,
                        }
            ), None, None, None)
        self.assertIn(obj.transfer_function_type,
                      ['LAPLACE (RADIANS/SECOND)', 'LAPLACE (HERTZ)',
                      'DIGITAL (Z-TRANSFORM)'],
                      f'transfer function type wrong in test case for PZ {obj.transfer_function_type}')
        self.assertEqual(obj.normalization_factor,  44188.013594177224,
                f'object normalization factor in test case for PZ {obj.normalization_factor} in PZ test is different from 44188.013594177224')
        print("1", end=" ")

        obj = Filter.construct(
            ObsMetadata({'type': 'PolesZeros',
                         'transfer_function_type' : 'LAPLACE (HERTZ)',
                         'zeros': ['0.3+0.2j'],
                         'poles': ['0.546+0.191j', '4.40e4+0.000j'],
                         'normalization_frequency' : 1.,
                         'normalization_factor' : None,
                        }
            ),
            None, None, None)
        self.assertIn(obj.transfer_function_type,
                      ['LAPLACE (RADIANS/SECOND)', 'LAPLACE (HERTZ)',
                       'DIGITAL (Z-TRANSFORM)'],
                      f'transfer function type wrong in test case for PZ {obj.transfer_function_type}')
        self.assertEqual(obj.normalization_factor,  50262.70428857582,
                         f'object normalization factor in test case for PZ {obj.normalization_factor} is different from 50262.70428857582')
        print("2", end=" ")

        obj = Filter.construct(
            ObsMetadata({'type': 'PolesZeros',
                         'transfer_function_type' : 'LAPLACE (RADIANS/SECOND)',
                         'zeros': ['0.3+0.2j'],
                         'poles': ['0.546+0.191j', '4.40e4+0.000j'],
                         'normalization_frequency' : 120.,
                         'normalization_factor' : None,
                        }
            ),
            None, None, None)

        self.assertEqual(obj.normalization_factor, 44006.99311749303, f'{obj.normalization_factor} in test case for PZ is different from 44006.99311749303')
        print("3", end=" ")

        obj = Filter.construct(
            ObsMetadata({'type': 'PolesZeros',
                         'transfer_function_type': 'LAPLACE (RADIANS/SECOND)',
                         'zeros': [],
                         'poles': [],
                         'normalization_frequency': 1.,
                         'normalization_factor': None
                        }
                       ),
            None, None, None)

        self.assertEqual(obj.normalization_factor, 1.,
                         f'{obj.normalization_factor} is different from 1.')
        print("4", end=" ")

        obj = Filter.construct(
            ObsMetadata({'type': 'PolesZeros',
                         'transfer_function_type': 'DIGITAL (Z-TRANSFORM)',
                         'zeros': ['0.3+0.2j'],
                         'poles': ['0.546+0.191j', '4.40e4+0.000j'],
                         'normalization_frequency': 1.,
                         'normalization_factor': None
                        }
            ),
            None, None, None)

        self.assertIn(obj.transfer_function_type,
                      ['LAPLACE (RADIANS/SECOND)', 'LAPLACE (HERTZ)',
                       'DIGITAL (Z-TRANSFORM)'],
                      'transfer function type wrong in PZ test case {}'
                      .format(obj.transfer_function_type))
        self.assertEqual(obj.normalization_factor, None,
                         'object normalization factor {} is not None'
                         .format(obj.normalization_factor))
        print("5", end=" ")

        obj = Filter.construct(
            ObsMetadata({'type': 'PolesZeros',
                         'transfer_function_type': 'LAPLACE (RADIANS/SECOND)',
                         'zeros': ['0.3+0.2j'],
                         'poles': ['0.546+0.191j', '4.40e4+0.000j'],
                         'normalization_frequency': None,
                         'normalization_factor': None,
                         }
                        ),
            None, None, None)

        self.assertEqual(obj.normalization_factor, None, f'{obj.normalization_factor} is different from None')
        print("6: Should have returned error", end=" ")

    def _test_all_stage_types(self):
        """
        Test reading and converting to obspy a stage file with each filter type.

        This is the first time obspy conversion occurs so make sure it's done right
        Only one example stage for each type. If you wish to test all stage files,
        use test_all_responses(). If you wish to test individual files use test_stage
        with file as stripped file name.
        File must exist in obsinfo/Information_files/XXX/responses
        where XXX = sensor, preamplifier or datalogger
        """
        self.test_stage('TI_ADS1281_FIR1.stage.yaml')
        self.test_stage('SIO-LDEO_DPG_5018_calibrated.stage.yaml')
        self.test_stage('test-with-coeff.stage.yaml')
        self.test_stage('test-with-response-list.stage.yaml')

    def _test_stage(self, file):
        """
        Test or print stage according to contained filter type

        :param: file - Filename to test or print
        :type file: str
        """
        if verbose:
            print(f'Processing stage file:"{file}"')

        info_file_dict = ObsMetadata.read_info_file(file, self.infofiles_path)
        stage_from_info_file = Stage(ObsMetadata(info_file_dict['stage']))
        obspy_result = stage_from_info_file.to_obspy()

        if self.test:
            self._test_common_attributes(stage_from_info_file, obspy_result)
            if isinstance(filter, FIR):
                self.assertEqual(stage_from_info_file.filter.symmetry, obspy_result._symmetry)
                for info_file_coeff in stage_from_info_file.filter.coefficients:
                    for obspy_coeff in obspy_result.decimation_correction:
                        self.assertEqual(info_file_coeff / 512, obspy_coeff(f))
            elif isinstance(filter, PolesZeros):
                self.assertEqual(stage_from_info_file.filter.transfer_function_type, obspy_result.pz_transfer_function_type)
                self.assertEqual(stage_from_info_file.filter.normalization_frequency, obspy_result.normalization_frequency)
                self.assertEqual(stage_from_info_file.filter.normalization_factor, obspy_result.normalization_factor)
                self.assertEqual(stage_from_info_file.filter.zeros, obspy_result.zeros)
                self.assertEqual(stage_from_info_file.filter.poles, obspy_result.poles)
            elif isinstance(filter, ResponseList):
                self.assertEqual(stage_from_info_file.filter.response_list, obspy_result.response_list_elements)
            elif isinstance(filter, Coefficients):
                self.test_common_attributes(stage_from_info_file, obspy_result)
                self.assertEqual(stage_from_info_file.filter.transfer_function_type, obspy_result.cf_transfer_function_type)
                self.assertEqual(stage_from_info_file.filter.numerator_coefficients, obspy_result.numerator)
                self.assertEqual(stage_from_info_file.filter.denominator_coefficients, obspy_result.denominator)
        if verbose:
            print(f'Stage test for: {file}: PASSED')
        if self.print_output:
            print(self)
            if self.level == "all":
                print(self.filter)

    def _test_common_attributes(self, stage_from_info_file, obspy_result):
        """
        Test attributes common to all stages

        :param stage_from_info_file:  Stage portion of dictionary with attributes
        :type stage_from_info_file: object of class:`Stage`
        :param obspy_result: Dictionary generated by obspy with corresponding attributes
        :type obspy_result: object of class ``Stage`` in ``osbpy.core.inventory.response``
        """
        self.assertEqual(stage_from_info_file.name, obspy_result.name)
        self.assertEqual(stage_from_info_file.description, obspy_result.description)
        self.assertEqual(stage_from_info_file.input_units, obspy_result.input_units)
        self.assertEqual(stage_from_info_file.output_units, obspy_result.output_units)
        self.assertEqual(stage_from_info_file.input_units_description, obspy_result.input_units_description)
        self.assertEqual(stage_from_info_file.output_units_description, obspy_result.output_units_description)
        self.assertEqual(stage_from_info_file.gain, obspy_result.stage_gain)
        self.assertEqual(stage_from_info_file.gain_frequency, obspy_result.stage_gain_frequency)
        self.assertEqual(stage_from_info_file.decimation_factor, obspy_result.decimation_factor)
        self.assertEqual(stage_from_info_file.filter.offset, obspy_result.decimation_offset)
        self.assertEqual(stage_from_info_file.delay, obspy_result.decimation_delay)
        self.assertEqual(stage_from_info_file.correction, obspy_result.decimation_correction)

    def _test_response_stage_addition(self):
        """
        Test reading and combining stages from a sensor and a datalogger
        """
        read_info_A = ObsMetadata.read_info_file(PurePath(self.infofiles_path).joinpath(
            'sensors',
            'responses',
            'Trillium_T240_SN400-singlesided_theoretical.stage.yaml'))
        read_info_B = ObsMetadata.read_info_file(PurePath(self.infofiles_path).joinpath(
            'dataloggers',
            'responses',
            'TexasInstruments_ADS1281_100sps-linear_theoretical.stages.yaml'))
        stages_A = Stages(read_info_A['response']['stages'])
        stages_B = Stages(read_info_B['response']['stages'])
        stages = stages_A + stages_B

    def _test_equipment_attributes(self, equipment_from_info_file, obspy_result):
        """
        Tesl the equipment portion of a component or instrumentation

        :param: equipment_from_info_file - Stage portion of dictionary with attributes
        :type equipment:  object of class:`Stage`
        :param: obspy_result - Dictionary generated by obspy with corresponding attributes
        :type obspy_result: object of class ``Stage`` in ``osbpy.core.inventory.response`
        """
        try:
            self.assertEqual(equipment_from_info_file.type, obspy_result.type)
            self.assertEqual(equipment_from_info_file.description,
                             obspy_result.description)
            self.assertEqual(equipment_from_info_file.manufacturer,
                             obspy_result.manufacturer)
            self.assertEqual(equipment_from_info_file.model,
                             obspy_result.model)
            self.assertEqual(equipment_from_info_file.vendor,
                             obspy_result.vendor)
            self.assertEqual(equipment_from_info_file.serial_number,
                             obspy_result.serial_number)
            self.assertEqual(
                UTCDateTime(equipment_from_info_file.installation_date)
                if equipment_from_info_file.installation_date else None,
                obspy_result.installation_date)
            self.assertEqual(
                UTCDateTime(equipment_from_info_file.removal_date)
                if equipment_from_info_file.removal_date else None,
                obspy_result.removal_date)
            for dt, obspy_dt in zip(equipment_from_info_file.calibration_dates, obspy_result.calibration_dates):
                self.assertEqual(UTCDateTime(dt) if dt else None, obspy_dt)
            self.assertEqual(equipment_from_info_file.resource_id, obspy_result.resource_id)
        except TypeError:
            print("TypeError, probably in UTCDateTime conversion")

    def test_sensor_configurations(self):
        """
        Test cases for configurations of particular sensor examples

        Will test what happens if no default is specified, if a default is
        specified and if default is overridden

        WARNING: Depends critically on the expected values of information
        files. If these change, tests will fail.
        """
        warnings.warn('Not testing local information files')
        return
        component_dir = Path(__file__).parent.joinpath(
            "data", "Information_Files", "components")

        sensor_wo_configs = str(component_dir.joinpath("HITECH_HTI04-PLC-ULF-wo-configs.sensor.yaml").resolve())
        sensor_w_default = str(component_dir.joinpath("NANOMETRICS_T240_w_config_and_default.sensor.yaml").resolve())
        sensor_w_configs = str(component_dir.joinpath("NANOMETRICS_T240_w_config_no_default.sensor.yaml").resolve())

        # expected value for all test cases
        expected_gain_value_1 = 0.000195
        expected_gain_value_2 = 598.45
        expected_gain_value_3 = 594.5

        if verbose:
            print(f'Processing sensor file w/o configurations: {sensor_wo_configs}')

        # Case 1;
        # Expect message :"No configuration key or default found in sensor. Configurations, if present, will not be applied

        obj = self._read_and_get_component(sensor_wo_configs, "sensor")
        self.assertEqual(obj.stages[0].gain, expected_gain_value_1)

        if verbose:
            print(f'Processing sensor file with default and configurations: {sensor_w_default}')

        # Case 2;
        # Expect default to be applied as no other configuration is specified
        # Assumes configuration_default: "SINGLE-SIDED_SN1-399"
        obj = self._read_and_get_component(sensor_w_default, "sensor")
        self.assertEqual(obj.stages[0].gain, expected_gain_value_2)

        # Expect default to be overridden by selection specified
        obj = self._read_and_get_component(
            sensor_w_default, "sensor",
            config_selector="SINGLE-SIDED_SN400plus")
        self.assertEqual(obj.stages[0].gain, expected_gain_value_3)

        if verbose:
            print(f'Processing sensor file with configurations but no default: {sensor_w_configs}')

        # Case 3:
        # Expect selected configuration selection to be applied as there is no default
        logging.disable()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            obj = self._read_and_get_component(
                sensor_w_configs, "sensor",
                config_selector="SINGLE-SIDED_SN400plus")
            self.assertEqual(obj.stages[0].gain, expected_gain_value_3)

            # Expect default to be overridden
            obj = self._read_and_get_component(
                sensor_w_configs, "sensor",
                config_selector="SINGLE-SIDED_SN400plus")
            self.assertEqual(obj.stages[0].gain, expected_gain_value_3)

            # Expect not to find configuration specified and no default and
            # thus raise TypeError
            logging.disable()   # Don't log this error
            self.assertRaises(TypeError, self._read_and_get_component,
                              sensor_w_configs, "sensor",
                              config_selector="SGLE-SIDED_SN1-399")

            # This will raise an exception as there are no stages due to lack of selection
            # self.assertRaises(TypeError, self._read_and_get_component,
            #                   sensor_w_configs, "sensor")
            logging.disable(logging.NOTSET)  # Turns normal logging back on

    def test_preamp_configurations(self):
        """
        Test all configurations of particular preamp examples

        Will not test case 1 in test_sensor_configurations as it is common
        code. Will test what happens if configuration is not
        found or is empty.

        WARNING: Depends critically on the expected values of information
        files. If these change, tests will fail.
        """
        warnings.warn('Not testing local information files')
        return
        component_dir = Path(__file__).parent.joinpath("data",
                                                       "Information_Files",
                                                       "components")

        preamp_w_configs = str(component_dir.joinpath("LCHEAPO_HYDRO.preamplifier.yaml").resolve())
        preamp_wo_default = str(component_dir.joinpath("LCHEAPO_HYDRO_wo_default.preamplifier.yaml").resolve())
        preamp_w_empty_config = str(component_dir.joinpath("LCHEAPO_GEOPHONE.preamplifier.yaml").resolve())
        # preamp_w_empty_config_no_stages = str(component_dir.joinpath("LCHEAPO_GEOPHONE-no-stages.preamplifier.yaml").resolve())

        # expected value for all test cases
        expected_gain_mult_1 = 16
        expected_gain_mult_2 = 32
        expected_gain_mult_3 = 64
        expected_gain_mult_4 = 128

        if verbose:
            print(f'Processing preamplifier file with configurations: {preamp_w_configs}')

        # Case 1;
        # Select different configs over default
        obj = self._read_and_get_component(preamp_w_configs, "preamplifier", config_selector="16x gain")
        self.assertEqual(obj.stages[0].gain, expected_gain_mult_1)

        obj = self._read_and_get_component(preamp_w_configs, "preamplifier", config_selector="32x gain")
        self.assertEqual(obj.stages[0].gain, expected_gain_mult_2)

        obj = self._read_and_get_component(preamp_w_configs, "preamplifier", config_selector="64x gain")
        self.assertEqual(obj.stages[0].gain, expected_gain_mult_3)

        obj = self._read_and_get_component(preamp_w_configs, "preamplifier", config_selector="128x gain")
        self.assertEqual(obj.stages[0].gain, expected_gain_mult_4)

        if verbose:
            print(f'Processing preamplifier file without default: {preamp_wo_default}')

        # Case 2:
        # select configuration without default
        logging.disable()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            obj = self._read_and_get_component(preamp_wo_default, "preamplifier", config_selector="128x gain")
            self.assertEqual(obj.stages[0].gain, expected_gain_mult_4)

            # Case 3:
            # No select configuration, no  default config
            # Since a default stages is specified, it should be used
            obj = self._read_and_get_component(preamp_wo_default, "preamplifier")
            self.assertEqual(obj.stages[0].gain, expected_gain_mult_3)

            if verbose:
                print(f'Processing preamplifier file with empty config: {preamp_w_empty_config}')

            # Case 4:
            # Config cannot be found. Expect error
            self.assertRaises(TypeError, self._read_and_get_component,
                              preamp_w_empty_config, "preamplifier",
                              config_selector="16x gain")
        logging.disable(logging.NOTSET)

    def test_datalogger_configurations(self):
        """
        Test all configurations of a particular datalogger example

        Will not test case 1 in test_sensor_configurations as it is common
        code. Will test what happens if configuration is not
        found or is empty.

        WARNING: Depends critically on the expected values of information
        files. If these change, tests will fail.
        """
        warnings.warn('Not testing local information files')
        return
        component_dir = Path(__file__).parent.joinpath(
            "data", "Information_Files", "components")

        datalogger_w_configs = str(component_dir.joinpath("LC2000.datalogger.yaml").resolve())

        # expected value for all test cases
        expected_sample_rate_1 = 62.5
        expected_sample_rate_2 = 125
        expected_sample_rate_3 = 250
        expected_sample_rate_4 = 500
        expected_sample_rate_5 = 1000

        if verbose:
            print(f'Processing datalogger file with configurations: {datalogger_w_configs}')

        # Case 1;
        # Select different configs over default
        obj = self._read_and_get_component(datalogger_w_configs, "datalogger",
                                           config_selector="62.5sps")
        self.assertEqual(self._calc_sample_rate(obj.stages),
                         expected_sample_rate_1)

        obj = self._read_and_get_component(datalogger_w_configs, "datalogger",
                                           config_selector="125sps")
        self.assertEqual(self._calc_sample_rate(obj.stages),
                         expected_sample_rate_2)

        obj = self._read_and_get_component(datalogger_w_configs, "datalogger",
                                           config_selector="250sps")
        self.assertEqual(self._calc_sample_rate(obj.stages),
                         expected_sample_rate_3)

        obj = self._read_and_get_component(datalogger_w_configs, "datalogger",
                                           config_selector="500sps")
        self.assertEqual(self._calc_sample_rate(obj.stages),
                         expected_sample_rate_4)

        obj = self._read_and_get_component(datalogger_w_configs, "datalogger",
                                           config_selector="1000sps")
        self.assertEqual(self._calc_sample_rate(obj.stages),
                         expected_sample_rate_5)

    def _calc_sample_rate(self, stage_list):
        """
        Calculates the total sample rate for the response

        using a different method than the actual program

        Args:
            stage_list (list of :class:`.Stage`): stages with input_sample_rate
                and decimation_factor

        Returns:
            the calculated sample rate
        """
        sample_rate = stage_list[0].input_sample_rate
        for st in stage_list[1:]:
            sample_rate /= st.decimation_factor
        return sample_rate

    def _read_and_get_component(self, file, component, channel_modif={},
                                correction=None, config_selector=''):
        """
        Read a component information file and create corresponding *obsinfo* object

        Args:
            file (str): filename of information file with complete path
            component (str): type of component
        Returns:
            (:class:`.Information_Component`)
        """
        info_file_dict = ObsMetadata.read_info_file(file, self.infofiles_path)
        obj = InstrumentComponent.dynamic_class_constructor(
            component, info_file_dict, config_selector=config_selector)
        return obj

    def _test_instrumentation(self, file):
        """
        Test or print a single instrumentation file

        WARNING: Response is assumed to be checked stage by stage at the stage level (test_stage),
        Here some trivial checks are done, such as number of stages
        SENSITIVITY MUST BE CHECKED MANUALLY in the StationXML file

        Args:
            file (str): Filename to test or print
        """
        if verbose:
            print(f'instrumentation file:"{file}"', end=" ", flush=True)

        dict = ObsMetadata.read_info_file(file, self.infofiles_path)

        start_date = end_date = "2021-01-03"

        # create a dummy location dictionary for testing
        location_dict = {
            "00": {
                "position": {"lon": 0., "lat": 0., "elev": 200.},
                "base": {
                    "depth.m": 100.,
                    "geology": "unknown",
                    "vault": "Sea floor",
                    "uncertainties.m": {"lon": 1., "lat": 1., "elev": 1.},
                    "localisation_method": "Sea surface release point",
                }
            },
            "01": {
                "position": {"lon": 0., "lat": 0., "elev": 300.},
                "base": {
                    "depth.m": 200.,
                    "geology": "unknown",
                    "vault": "Sea floor",
                    "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                    "localisation_method": "Sea surface release point"
                }
            },
            "02": {
                "position": {"lon": 0., "lat": 0., "elev": 300.},
                "base": {
                    "depth.m": 200.,
                    "geology": "unknown",
                    "vault": "Sea floor",
                    "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                    "localisation_method": "Sea surface release point"
                }
            },
            "03": {
                "position": {"lon": 0., "lat": 0., "elev": 300.},
                "base": {
                    "depth.m": 200.,
                    "geology": "unknown",
                    "vault": "Sea floor",
                    "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                    "localisation_method": "Sea surface release point"
                }
            },
            "04": {
                "position": {"lon": 0., "lat": 0., "elev": 300.},
                "base": {
                    "depth.m": 200.,
                    "geology": "unknown",
                    "vault": "Sea floor",
                    "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                    "localisation_method": "Sea surface release point"
                }
            },
            "05": {
                "position": {"lon": 0., "lat": 0., "elev": 300.},
                "base": {
                    "depth.m": 200.,
                    "geology": "unknown",
                    "vault": "Sea floor",
                    "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                    "localisation_method": "Sea surface release point"
                }
            },
            "06": {
                "position": {"lon": 0., "lat": 0., "elev": 300.},
                "base": {
                    "depth.m": 200.,
                    "geology": "unknown",
                    "vault": "Sea floor",
                    "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                    "localisation_method": "Sea surface release point"
                }
            },
            "07": {
                "position": {"lon": 0., "lat": 0., "elev": 300.},
                "base": {
                    "depth.m": 200.,
                    "geology": "unknown",
                    "vault": "Sea floor",
                    "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                    "localisation_method": "Sea surface release point"
                }
            },
            "08": {
                "position": {"lon": 0., "lat": 0., "elev": 300.},
                "base": {
                    "depth.m": 200.,
                    "geology": "unknown",
                    "vault": "Sea floor",
                    "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                    "localisation_method": "Sea surface release point"
                }
            },
            "09": {
                "position": {"lon": 0., "lat": 0., "elev": 300.},
                "base": {
                    "depth.m": 200.,
                    "geology": "unknown",
                    "vault": "Sea floor",
                    "uncertainties.m": {"lon": 2., "lat": 2., "elev": 2.},
                    "localisation_method": "Sea surface release point"
                }
            },
        }

        locations = {c: Location(v) for c, v in location_dict.items()}

        obj = Instrumentation(ObsMetadata(dict['instrumentation']), locations,
                              start_date, end_date, {})

        if self.test:
            for ch in obj.channels:

                self.assertEqual(ch.channel_code(ch.instrument.sample_rate),
                                 ch.obspy_channel.code)
                self.assertEqual(ch.location_code,
                                 ch.obspy_channel.location_code)
                self.assertEqual(ch.location.obspy_latitude,
                                 ch.obspy_channel.latitude)
                self.assertEqual(ch.location.obspy_longitude,
                                 ch.obspy_channel.latitude)
                self.assertEqual(ch.location.elevation,
                                 ch.obspy_channel.elevation)
                self.assertEqual(ch.location.depth_m, ch.obspy_channel.depth)
                self.assertEqual(ch.orientation.azimuth,
                                 ch.obspy_channel.azimuth)
                self.assertEqual(ch.orientation.dip, ch.obspy_channel.dip)
                self.assertEqual(ch.instrument.sample_rate,
                                 ch.obspy_channel.sample_rate)
                self.assertEqual(ch.instrument.sensor.obspy_equipment,
                                 ch.obspy_channel.sensor)
                self.assertEqual(ch.instrument.datalogger.obspy_equipment,
                                 ch.obspy_channel.data_logger)
                preamp = (ch.instrument.preamplifier.obspy_equipment
                          if ch.instrument.preamplifier else None)
                self.assertEqual(preamp, ch.obspy_channel.pre_amplifier)
                self.assertEqual(UTCDateTime(ch.start_date) if ch.start_date
                                 else None,
                                 ch.obspy_channel.start_date)
                self.assertEqual(UTCDateTime(ch.end_date) if ch.end_date
                                 else None,
                                 ch.obspy_channel.end_date)
                # self.assertEqual(ch.channel_id_code,
                #                  ch.obspy_channel.description)
                self.assertEqual(len(ch.instrument.stages),
                                 len(ch.obspy_channel.response.stages))

                self._test_equipment_attributes(obj.equipment,
                                                obj.equipment.obspy_equipment)
        if verbose:
            print('PASSED')

        if self.print_output:
            PrintObs.print_instrumentation(obj, self.level)

    def _test_station(self, file_name="", info_dict={}, read_file=True):
        """
        Test or print a station.

        Args:
            file_name (str): Filename to test or print
            info_dict (dict or :class:``.ObsMetadata``): If not reading
                file, MUST provide an info_dict with the info
            read_file (bool): indicates whether file should be read or
                info_dict will be provided

        WARNING: Comments, extras and Processing string must be checked visually
        WARNING: Check operator visually
        """
        if self.test and not read_file and not info_dict:
            print("MUST provide info_dict if read_file is False")

        if read_file:
            info_dict = ObsMetadata.read_info_file(file_name,
                                                   self.infofiles_path)
            info_dict = info_dict['station']

        key = list(info_dict.keys())[0]
        value = ObsMetadata(list(info_dict.values())[0])

        try:
            obj = Station(key, value)
            station_only = False
        except Exception:
            obj = Station(key, value, station_only=True)
            station_only = True

        if self.test:
            # latitude, longitude = Location.get_obspy_latitude_and_longitude(obj.location)
            site = Site(name=obj.site, description=None, town=None, county=None, region=None, country=None)
            self.assertEqual(obj.label, obj.obspy_station.code)
            self.assertEqual(site.name, obj.obspy_station.site.name)
            self.assertEqual(UTCDateTime(obj.start_date) if obj.start_date else None,
                             obj.obspy_station.creation_date)
            self.assertEqual(UTCDateTime(obj.end_date) if obj.end_date else None,
                             obj.obspy_station.termination_date)
            self.assertEqual(obj.restricted_status,
                             obj.obspy_station.restricted_status)
            self.assertEqual(obj.location.obspy_latitude,
                             obj.obspy_station.latitude)
            self.assertEqual(obj.location.obspy_longitude,
                             obj.obspy_station.longitude)
            self.assertEqual(obj.location.elevation, obj.obspy_station.elevation)
            self.assertEqual(obj.location.vault, obj.obspy_station.vault)
            self.assertEqual(obj.location.geology, obj.obspy_station.geology)

            if station_only is False:
                # Check if locations are correctly assigned to channels
                for ch in obj.instrumentation.channels:
                    try:
                        self.assertEqual(obj.locations[ch.location_code].latitude, ch.location.latitude)
                    except TypeError:
                        pass

                # Check number of channels OK
                num_channels = len(obj.instrumentation.channels)
                self.assertEqual(num_channels, len(obj.obspy_station.channels))
                self.assertEqual(num_channels,
                                 obj.obspy_station.total_number_of_channels)
                self.assertEqual(num_channels,
                                 obj.obspy_station.selected_number_of_channels)

        if verbose:
            print(f'Processing station file: {file_name}: PASSED')

        if self.print_output:
            PrintObs.print_station(obj, self.level)

    def _test_subnetwork(self, file_name):
        """
        Test or print a subnetwork information file

        :param: file_name - Filename to test or print
        :type file_name: str

        WARNING: Check operator visually
        """
        info_dict = ObsMetadata.read_info_file(file_name, self.infofiles_path)

        subnet_dict = info_dict.get('subnetwork', None)
        if not subnet_dict:
            return

        if verbose:
            print(f'Processing subnetwork file: {file_name}')

        try:
            logging.disable()
            obj = Subnetwork(ObsMetadata(subnet_dict))
            logging.disable(logging.NOTSET)
        except Exception:
            obj = Subnetwork(ObsMetadata(subnet_dict), station_only=True)

        if self.test:
            # description = obj.fdsn_name + " -" + obj.description
            self.assertEqual(obj.fdsn_code, obj.obspy_network.code)
            self.assertEqual(UTCDateTime(obj.start_date)
                             if obj.start_date else None,
                             obj.obspy_network.start_date)
            self.assertEqual(UTCDateTime(obj.end_date)
                             if obj.end_date else None,
                             obj.obspy_network.end_date)
            self.assertEqual(obj.restricted_status,
                             obj.obspy_network.restricted_status)

            # Check number of channels OK
            num_stations = len(obj.stations)
            self.assertEqual(num_stations, len(obj.obspy_network.stations))
            self.assertEqual(num_stations,
                             obj.obspy_network.total_number_of_stations)
            self.assertEqual(num_stations,
                             obj.obspy_network.selected_number_of_stations)

            self._test_station(info_dict=subnet_dict['stations'], read_file=False)

        if verbose:
            print(f'Subnetwork test for: {file_name}: PASSED')

        if self.print_output:
            PrintObs.print_subnetwork(obj, self.level)

    def test_change_modifications(self):
        """
        Test all possible channel modification and response modification labels

        Uses several test subnetwork information files plus a test instrumentation
        file in _examples. Rationale for each assertEqual is given in the test
        information files themselves

        The five first tests check channel modification labels and their
        priorities. The last one checks stage_modifications

        WARNING: Depends critically on the expected values of information
        files. If these change, tests will fail.

         * channel 0 - 3-00
         * channel 1 - H-00
         * channel 2 - H-01
         * channel 3 - 3-02
        """
        expected_sample_rate = 62.5
        expected_sample_rate_2 = 500
        expected_sample_rate_3 = 250
        unchanged_sample_rate = 125

        proc_str_chan = "Processing subnetwork file with channel modifications"
        proc_str_stage = "Processing subnetwork file with stage modifications"

        if verbose:
            print('Processing subnetwork file with channel modifications: TEST-channel-mods-1.subnetwork.yaml')

        obj = self._create_test_subnetwork("TEST-channel-mods-1.subnetwork.yaml")
        inst = obj.stations[0].instrumentation

        self.assertEqual(inst.channels[0].instrument.sample_rate, expected_sample_rate)
        self.assertEqual(inst.channels[1].instrument.sample_rate, expected_sample_rate)
        self.assertEqual(inst.channels[2].instrument.sample_rate, unchanged_sample_rate)
        self.assertEqual(inst.channels[3].instrument.sample_rate, unchanged_sample_rate)

        if verbose:
            print(f'{proc_str_chan}: TEST-channel-mods-2.subnetwork.yaml')

        obj = self._create_test_subnetwork("TEST-channel-mods-2.subnetwork.yaml")
        inst = obj.stations[0].instrumentation

        self.assertEqual(inst.channels[0].instrument.sample_rate,
                         expected_sample_rate)
        self.assertEqual(inst.channels[1].instrument.sample_rate,
                         expected_sample_rate)
        self.assertEqual(inst.channels[2].instrument.sample_rate,
                         expected_sample_rate_2)
        self.assertEqual(inst.channels[3].instrument.sample_rate,
                         expected_sample_rate_3)

        if verbose:
            print(f'{proc_str_chan}: TEST-channel-mods-3.subnetwork.yaml')

        obj = self._create_test_subnetwork("TEST-channel-mods-3.subnetwork.yaml")
        inst = obj.stations[0].instrumentation

        self.assertEqual(inst.channels[0].instrument.sample_rate, expected_sample_rate_2)
        self.assertEqual(inst.channels[1].instrument.sample_rate, expected_sample_rate)
        self.assertEqual(inst.channels[2].instrument.sample_rate, expected_sample_rate)
        self.assertEqual(inst.channels[3].instrument.sample_rate, expected_sample_rate_3)

        if verbose:
            print(f'{proc_str_chan}: TEST-channel-mods-4.subnetwork.yaml')

        obj = self._create_test_subnetwork("TEST-channel-mods-4.subnetwork.yaml")
        inst = obj.stations[0].instrumentation

        self.assertEqual(inst.channels[0].instrument.sample_rate, expected_sample_rate)
        self.assertEqual(inst.channels[1].instrument.sample_rate, expected_sample_rate)
        self.assertEqual(inst.channels[2].instrument.sample_rate, expected_sample_rate)
        self.assertEqual(inst.channels[3].instrument.sample_rate, expected_sample_rate)

        if verbose:
            print(f'{proc_str_chan}: TEST-channel-mods-5.subnetwork.yaml')

        obj = self._create_test_subnetwork("TEST-channel-mods-5.subnetwork.yaml")
        inst = obj.stations[0].instrumentation

        self.assertEqual(inst.channels[0].instrument.sample_rate, expected_sample_rate_3)
        self.assertEqual(inst.channels[1].instrument.sample_rate, expected_sample_rate)
        self.assertEqual(inst.channels[2].instrument.sample_rate, expected_sample_rate)
        self.assertEqual(inst.channels[3].instrument.sample_rate, expected_sample_rate_3)

        # check response stages
        expected_gain = 123456.
        expected_coeff_1 = [0, .1, .2, .4]
        expected_symm_1 = "ODD"
        expected_coeff_2 = [5, 6, 4, 6]
        expected_symm_2 = "EVEN"
        expected_descr = "Labeled modifications"

        if verbose:
            print(f'{proc_str_stage}: TEST-stage-mods-1.subnetwork.yaml')

        obj = self._create_test_subnetwork("TEST-stage-mods-1.subnetwork.yaml")
        inst = obj.stations[0].instrumentation

        ins = inst.channels[0].instrument  # channel 3-00

        self.assertEqual(ins.sample_rate, expected_sample_rate)
        self.assertEqual(ins.sensor.stages[0].gain, expected_gain)
        # just check a few, all should be expected_gain
        for stage in ins.datalogger.stages:
            self.assertEqual(stage.gain, expected_gain)
        # check all in interval
        self.assertEqual(ins.datalogger.stages[5].description, expected_descr)
        self.assertEqual(ins.datalogger.stages[6].description, expected_descr)
        self.assertEqual(ins.datalogger.stages[7].description, expected_descr)
        self.assertEqual(ins.datalogger.stages[8].description, expected_descr)

        ins = inst.channels[1].instrument  # Channel H-00

        self.assertEqual(ins.sample_rate, expected_sample_rate)
        self.assertEqual(ins.datalogger.stages[3].gain, expected_gain)

        self.assertEqual(ins.datalogger.stages[4].filter.symmetry,
                         expected_symm_1)
        self.assertEqual(ins.datalogger.stages[4].filter.coefficients,
                         expected_coeff_1)
        self.assertEqual(ins.datalogger.stages[5].filter.symmetry,
                         expected_symm_2)
        self.assertEqual(ins.datalogger.stages[5].filter.coefficients,
                         expected_coeff_2)
        self.assertEqual(ins.datalogger.stages[6].filter.symmetry,
                         expected_symm_2)
        self.assertEqual(ins.datalogger.stages[6].filter.coefficients,
                         expected_coeff_2)

    def _create_test_subnetwork(self, subnetwork_file):
        """
        Read an information file for a subnetwork and return the corresponding dictionary

        :param subnetwork_file: file name to read and create
        :type subnetwork_file: str
        :returns: object of :class:`Subnetwork`
        """
        warnings.warn('Not testing local information files')
        return
        subnetwork_dir = Path(__file__).parent.joinpath(
            "data", "Information_Files", "subnetwork")

        subnetwork_file = str(subnetwork_dir.joinpath(subnetwork_file).resolve())

        info_dict = ObsMetadata.read_info_file(subnetwork_file, self.infofiles_path)

        return Subnetwork(ObsMetadata(info_dict.get('subnetwork', None)))


def run_suite_info_files(argv=None):
    """
    Create all test suites for information files
    """
    def suite():
        """ test suite """
        suite_info_files = unittest.TestSuite()

        suite_info_files.addTest(TestObsinfo('test_test_filters'))
        suite_info_files.addTest(TestObsinfo('test_sensor_configurations'))
        suite_info_files.addTest(TestObsinfo('test_preamp_configurations'))
        suite_info_files.addTest(TestObsinfo('test_datalogger_configurations'))
        suite_info_files.addTest(TestObsinfo('test_change_modifications'))

        return suite_info_files

    result = unittest.TextTestRunner(verbosity=1).run(suite())
    report_result_summary(result)


def report_result_summary(result):
    """
    Report a summary of errors and failures

    :param: result - Contains result stats, errors and failures.
    :type result: object returned by unittest TextTestRunner
    """
    n_errors = len(result.errors)
    n_failures = len(result.failures)

    if n_errors or n_failures:
        print(f'\n\nSummary: {n_errors} errors and {n_failures} failures reported\n')


if __name__ == '__main__':
    run_suite_info_files(["--all"])



"""Dump network/instrument info to standardized A-node file"""


def infodump(netinfo):
    """
    Args:
        netinfo (`obsinfo:Subetwork`): the network to dump

    Experiment > Campaign > Facility

    Dans la hiérarchie d'informations, le lien entre un "entity" enfant et un
    "entity" parent se fait au travers de la référence en utilisant la
    convention : "<child-key>@<parent-key>".

    Pour les liens avec des "entities" ne se trouvant pas dans la hiérarchie,
    je me suis inspiré des concepts de RDF (Resource Desciption Framework) :

    subject --predicate-> object

    Dans ce qui suit, on a :

    - sujet : une campagne (avec sa référence globalement unique)
    - objet : un réseau (identifié par sa référence unique)
    - prédicat : "est associé à" (forme verbale)

      "campaigns": [ {
          "_campaign-ref": "MOMAR_2007-2008_A@MOMAR",
          "name": "MOMAR_2007-2008_A",
          "is-associated-to": { "_network-ref": "4G+2007" },
    ...
        } ],

    - sujet : un parc OBS (avec sa référence globalement unique)
    - objet : des stations (identifiées par leur référence unique)
    - prédicat : "a opéré" (forme verbale)

      "facilities": [ {
          "_facility-ref": "INSU-IPGP@MOMAR_2007-2008_A@MOMAR",
          "name": "INSU-IPGP",
          "full-name": "INSU-IPGP OBS Park",
          "has-operated": { "_station-refs": [ "AZBBA@4G+2007",
                                               "LSV5A@4G+2007" ] }
        } ]
    """
    net_code = netinfo.network_code
    campaign_code = netinfo.campaign_ref
    facility_code = netinfo.operator.ref_name

    network_dict = dict(_network_ref=net_code,
                        fdsn_code=net_code,
                        name=netinfo.network_name,
                        operating_period=dict(start_date=netinfo.start_date,
                                              end_date=netinfo.end_date),
                        _campaign_ref=campaign_code)
    station_list = []
    for station in netinfo.stations:
        sta_code = station.code
        station_list.append(
            dict(_network_idx=0,
                 _station_ref=f'{sta_code}@{net_code}',
                 code=sta_code,
                 site_description=station.site,
                 aquisition_period=dict(start_datetime=station.start_time,
                                        end_datetime=station.end_time),
                 _facility_ref=f'{facility_code}@{campaign_code}'))



"""
Filter class and subclasses
"""
import logging

from .filter_template import FilterTemplate
from ...helpers import str_indent

logger = logging.getLogger("obsinfo")


class ResponseList(FilterTemplate):
    """
    ResponseList Filter

    Attributes:
        elements (list of lists): list of responses instead of function
            coefficients.  Each response is [freq, value, phase (deg)]
    """

    def __init__(self, attributes_dict, stage_id=-1):
        """
        Constructor

        Args:
            attributes_dict (dict or :class:`.ObsMetadata`): filter attributes
            stage_id (int): id of corresponding stage. Used for reporting only
        """
        logger.debug(f'in {self.__class__.__name__}.__init__()')
        super().__init__(attributes_dict, stage_id)
        self.elements = attributes_dict.pop('elements', [])
        self._validate_empty_attributes_dict(attributes_dict)

        # Validate attributes
        for element in self.elements:
            if not len(element) == 3:
                raise ValueError(f'ResponseList element {element} does not have 3 values')

    def __repr__(self):
        return f'          ResponseList("{self.elements}")'

    def __str__(self, indent=0, n_subclasses=0):
        """
        Args:
            indent (int): number of extra characters to indent lines by
            n_subclasses (int): unused
        """
        if n_subclasses < 0:
            return f'{type(self)}'
        s = super().__str__() + '\n'
        s += f'    {len(self.elements)} elements'
        return str_indent(s, indent)



from pathlib import Path
import obsinfo


def main():
    file = Path(obsinfo.__file__).parent.joinpath("version.py")

    version = {}
    with open(file) as fp:
        exec(fp.read(), version)

    name = "obsinfo"
    version = version['__version__']

    print(name + " v" + version)



"""
Station Class
"""
# Standard library modules
import warnings
import logging

from obspy.core.inventory.util import Site as obspy_site

from ..helpers import str_indent

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Site(object):
    def __init__(self, name):
        self.name = name
        
    def __str__(self, indent=0, n_subclasses=0):
        return str_indent(f'Site: {self.name}', indent)
    
    def to_obspy(self):
        return obspy_site(name=self.name, description=None, town=None, county=None,
                    region=None, country=None)


"""
Network :class:, network to which the subnetwork belongs
"""
# Standard library modules
import logging

# obsinfo modules
from .operator import Operators
from ..helpers import (str_indent, str_list_str, verify_dict_is_empty,
                       OIDate, Identifiers, Comments)

logger = logging.getLogger("obsinfo")


class Network(object):
    """
    Equivalent to obspy/StationXML Network

    Attributes:
         code (str)
         name (str)
         operators (:class:`Operators`): operators
         start_date (str with date format): network start date
         end_date (str with date format): network end date
         description  (str): network description
         comments (list of str):
         restricted_status (str): 'open', 'closed', or 'partial'
         source_id (str): network-level data source identifier in URI format
         identifiers (list of str): persistent identifiers as URIs, scheme (prefix) will be extracted
    """

    def __init__(self, attributes_dict=None):
        """
        Constructor

        Args:
            attributes_dict (dict or :class:`.ObsMetadata`): dictionary from
                subnetwork info file
        Raises:
            (TypeError): if attributes_dict is empty
        """
        if not attributes_dict:
            msg = 'No {self.__class__.__name__} attributes'
            logger.error(msg)
            raise TypeError(msg)

        self.code = attributes_dict.pop("code", None)
        self.name = attributes_dict.pop("name", None)
        self.operators = Operators(attributes_dict.pop("operators", None))
        self.start_date = OIDate(attributes_dict.pop("start_date", None))
        self.end_date = OIDate(attributes_dict.pop("end_date", None))
        self.description = attributes_dict.pop("description", None)
        self.restricted_status = attributes_dict.pop("restricted_status", None)
        self.source_id = attributes_dict.pop("source_id", None)
        self.identifiers = Identifiers(attributes_dict.pop("identifiers", None))
        self.comments = Comments(attributes_dict.pop("comments", None))
        verify_dict_is_empty(attributes_dict)

    def __str__(self, indent=0, n_subclasses=0):
        if n_subclasses < 0:
            return f'{type(self)}'
        kwargs = {'indent': 4, 'n_subclasses': n_subclasses-1}
        s = f'{self.__class__.__name__}:\n'
        s += f'    code: {self.code}\n'
        s += f'    name: {self.name}\n'
        s += f'    operators: {self.operators.__str__(**kwargs)}\n'
        s += f'    start_date: {self.start_date}\n'
        s += f'    end_date: {self.end_date}\n'
        s += f'    description: {self.description}\n'
        s += f'    restricted_status: {self.restricted_status}'
        if self.source_id is not None:
            s += f'\n    source_id: {self.source_id}'
        if self.identifiers is not None:
            s += f'\n    identifiers: {self.identifiers.__str__(**kwargs)}'
        if len(self.comments) > 0:
            s += f'\n    comments: {str_list_str(self.comments, **kwargs)}'
        return str_indent(s, indent)
